
/*
** Asynchronous driver for Berkeley DB 4.0.14, R8 style.
**
** Copyright (c) 2002, Scott Lystig Fritchie.  All rights reserved.
** See the file "LICENSE" at the top of the source distribution for
** full license terms.
**
*/

#include	<stdio.h>
#include	<unistd.h>
#include	<stdarg.h>
#include	<errno.h>
#include	<time.h>
#include	<limits.h>
#include	<sys/param.h>
#include	<pthread.h>	/* For debugging only */

#include	<erl_driver.h>
#include	"berkeley_db_drv.h"
#include	"external.h"	/* XXX deleteme someday */
#include	<db.h>

#ifdef	DMALLOC
#include	<dmalloc.h>
#endif	/* DMALLOC */

/* callstate_t = asychronous call state */
typedef struct callstate {
    struct callstate	*next;
    int			cmd;		/* call/command # */
#define	XXXKEY	((void *) 1)
    unsigned int	*key;		/* Not really used right now! */
    void		(*invoke)(void *);
    void		(*free)(void *);
    unsigned short	xid;	/* Only used by as_control()-initiated funcs */
    /* input args ... struct just for grouping args */
    struct {
	DB_ENV		*dbenvp;
	DB		*dbp;
	char		dbnum;
	DB_TXN		*txnid;
	DBC		*cursorp;
	char		cursornum;
	char		*file;
	char		*database;
	char		*outfile;
	char		*key;
	char		*val;
	u_int32_t	keylen;
	u_int32_t	vallen;
	int		type;
	u_int32_t	flags;
	int		mode;
    } i;
    /* output results ... struct just for grouping args */
    int			status;
    struct {
	void		*ptr;	/* Generic use ptr */
	char		*key;
	char		*val;
	u_int32_t	keylen;
	u_int32_t	vallen;
    } o;
} callstate_t;

/* descriptor = general state for the port */
typedef struct {
    ErlDrvPort		port;
    unsigned short	nextxid;	/* Call txn ID for very async calls */
    /*
    ** efile keeps an 'invoke' func pointer in the descriptor to assist
    ** with handling (aparently) timeout situations.
    */
    DB_ENV		*dbenvp;
    DB			*dbp[BDB_MAX_TABLES];
    DB_TXN		*txnid[BDB_MAX_TXNS];
    DBC			*cursorp[BDB_MAX_CURSORS];
    int			cursor2tab[BDB_MAX_CURSORS];
} descriptor;

static ErlDrvTermData am_ok;
static ErlDrvTermData am_error;
static ErlDrvTermData am_einval;
static ErlDrvTermData am_enotty;
static ErlDrvTermData am_enomem;
static ErlDrvTermData am_unknown;

#define	DEBUGIO		stderr
static int _debug = 0;

ErlDrvEntry *driver_init(void *);
static int bdb_init(void);
static ErlDrvData bdb_start(ErlDrvPort port, char *command);
static void bdb_stop(ErlDrvData drv_data);
static void bdb_output(ErlDrvData drv_data, char *buf, int len);
static int bdb_control(ErlDrvData drv_data, unsigned int command, char *buf, int len, char **rbuf, int rlen);
static void bdb_outputv(ErlDrvData drv_data, ErlIOVec *ev);
static void bdb_ready_async(ErlDrvData drv_data, ErlDrvThreadData thread_data);

/* XXX where do I find prototypes for these? */
/* XXX Should I use driver_{alloc,realloc,free}() instead??? */
void *sys_alloc(size_t);
void *sys_realloc(void *, size_t);
void sys_free(void *);

/* Wrapper functions for driver_{alloc,realloc,free}_binary */
static void *driver_alloc_wrapper(size_t);
static void *driver_realloc_wrapper(void *, size_t);
static void driver_free_wrapper(void *);
static ErlDrvBinary *alloced_ptr2ErlDrvBinary(void *);

static int reply_ok(descriptor *desc);
static int reply_ok_num(descriptor *desc, u_int32_t num);
static int reply_ok_key_val(descriptor *desc, callstate_t *);
static int reply_error(descriptor *desc, int errnum);
static int reply_tag_ok(descriptor *desc, unsigned short tag);
static int reply_tag_error(descriptor *desc, unsigned short tag, int errnum);
static int debug(char *f, ...);
static void debug_errcall(const char *, char *);
static void free_data(void *data);
static void close_hangers_on(descriptor *);

static void invoke_db_env_create(void *data);
static void invoke_db_env_open(void *data);
static void invoke_db_env_close(void *data);
static void invoke_db_env_remove(void *data);
static void invoke_db_env_set_rpc_server(void *data);
static void invoke_db_create(void *data);
static void invoke_db_open(void *data);
static void invoke_db_open(void *data);
static void invoke_db_put(void *data);
static void invoke_db_get(void *data);
static void invoke_db_del(void *data);
static void invoke_db_close(void *data);
static void invoke_db_upgrade(void *data);
/* XXX static void invoke_db_stat(void *data); */
static void invoke_db_truncate(void *data);
static void invoke_db_verify(void *data);
static void invoke_db_sync(void *data);
static void invoke_db_remove(void *data);

static void invoke_db_cursor(void *data);
static void invoke_db_c_get(void *data);
static void invoke_db_c_put(void *data);
static void invoke_db_c_del(void *data);
static void invoke_db_c_dup(void *data);
/* XXX static void invoke_db_join(void *data); */
static void invoke_db_c_count(void *data);
static void invoke_db_c_close(void *data);

static void invoke_db_set_pagesize(void *data);
static void invoke_db_set_cachesize(void *data);
static void invoke_db_set_lorder(void *data);
static void invoke_db_set_flags(void *data);
static void invoke_db_set_bt_minkey(void *data);
static void invoke_db_set_h_ffactor(void *data);
static void invoke_db_set_h_nelem(void *data);
static void invoke_db_set_re_delim(void *data);
static void invoke_db_set_re_len(void *data);
static void invoke_db_set_re_pad(void *data);
static void invoke_db_set_re_source(void *data);

#ifdef	XXX_TODO
static void invoke_db_txn_begin(void *data);
static void invoke_db_txn_abort(void *data);
static void invoke_db_txn_commit(void *data);
static void invoke_db_txn_discard(void *data);
static void invoke_db_txn_id(void *data);
static void invoke_db_txn_prepare(void *data);
#endif	/* XXX_TODO */

/* XXX These library funcs should be moved to a more general place */
static int ev_forward_N(ErlIOVec *ev, int n, int *pp, int *qp);

ErlDrvEntry bdb_driver_entry = {
    bdb_init,		       	/* init */
    bdb_start,			/* start */
    bdb_stop,			/* stop */
    bdb_output,			/* output */
    NULL,			/* ready_input */
    NULL,			/* ready_output */
    "berkeley_db_drv",		/* driver_name */
    NULL,			/* finish */
    NULL,			/* handle */
    bdb_control,		/* control */
    NULL,			/* timeout */
    bdb_outputv,		/* outputv */
    bdb_ready_async,		/* ready_async */
    NULL,			/* flush */
    NULL			/* call */
};

ErlDrvEntry *
driver_init(void *handle)
{
    bdb_driver_entry.handle = handle;
    return &bdb_driver_entry;
}

static int
bdb_init(void)
{

    am_ok = driver_mk_atom("ok");
    am_error = driver_mk_atom("error");

    am_einval = driver_mk_atom("einval");
    am_enotty = driver_mk_atom("enotty");
    am_enomem = driver_mk_atom("enomem");
    am_unknown = driver_mk_atom("unknown");

    return 0;
}

static ErlDrvData
bdb_start(ErlDrvPort port, char *args)
{
    descriptor  *desc;
    int		i;

    debug("bdb_start: starting, port = %ld, args = 0x%lx, %s", port,
	  (unsigned long) args, args);

    if ((desc = (descriptor *) sys_alloc(sizeof(descriptor))) == NULL)
	return ERL_DRV_ERROR_GENERAL;
    memset(desc, 0, sizeof(descriptor));
    for (i = 0; i < BDB_MAX_CURSORS; i++) {
	desc->cursor2tab[i] = -1;
    }
    desc->port = port;
    desc->nextxid = 4242;
    /* Everything else taken care of by memset() */

#ifdef	XXX_BUG_IN_BEAM
    /*
    ** It would be cool if set_port_control_flags() worked in R8B-0 and
    ** R8B-1, but it doesn't, at least under FreeBSD 4.5.
    */
    set_port_control_flags(port, PORT_CONTROL_FLAG_BINARY);
#endif
    return (ErlDrvData) desc;
}

static void
bdb_stop(ErlDrvData drv_data)
{
    descriptor	*desc = (descriptor *) drv_data;
    int		i;

    if (desc == NULL) {
	debug("bdb_stop: drv_data == NULL");
	return;
   }
    debug("bdb_stop: port = %ld", desc->port);
    if (desc->dbenvp != NULL) {
	debug("%s: closing environment", __FUNCTION__, i);
	close_hangers_on(desc);
	desc->dbenvp->close(desc->dbenvp, 0);
    }
    sys_free(desc);    
}

static void
close_hangers_on(descriptor *desc)
{
    int	i;

    debug("%s: top", __FUNCTION__);
    /*
    ** TODO: Oi, this will be fun once we have first-class support for
    ** cursors & transactions: we can have all sorts of race conditions
    ** and deadlock possibilities since we should be closing all those
    ** things using the same thread that created them ... but some of
    ** them might be busy right now.  :-|
    */
    for (i = 0; i < BDB_MAX_CURSORS; i++) {
	if (desc->cursorp[i] != NULL) {
	    debug("%s: closing cursor at slot %d", __FUNCTION__, i);
	    desc->cursorp[i]->c_close(desc->cursorp[i]);
	    desc->cursorp[i] = NULL;
	    desc->cursor2tab[i] = -1;
	}
    }
    for (i = 0; i < BDB_MAX_TXNS; i++) {
	if (desc->txnid[i] != NULL) {
	    debug("%s: aborting txn at slot %d", __FUNCTION__, i);
	    desc->txnid[i]->abort(desc->txnid[i]);
	    desc->txnid[i] = NULL;
	}
    }
    for (i = 0; i < BDB_MAX_TABLES; i++) {
	if (desc->dbp[i] != NULL) {
	    debug("%s: closing table at slot %d", __FUNCTION__, i);
	    desc->dbp[i]->close(desc->dbp[i], 0);
	    desc->dbp[i] = NULL;
	}
    }
}

static void
bdb_output(ErlDrvData drv_data, char *buf, int len)
{
    debug("bdb_output: XXX someone tried to call us, silly");
}

static int
bdb_control(ErlDrvData drv_data, unsigned int command, char *buf, int len,
	   char **rbuf, int rlen)
{
    descriptor	*desc = (descriptor *) drv_data;
    callstate_t	*c = NULL;
    char	*ret = *rbuf;
    int		retlen;

    switch (command) {
#ifdef	XXX_ASLEEP_EXAMPLE
    case 0 /* XXX */:
	if (len != (2 * sizeof(unsigned long))) {
	    debug("bdb_control: asleep: bad len = %d", len);
	    ret[0] = 1;		/* XXX */
	    retlen = 1;
	} else {
	    if ((c = sys_alloc(sizeof(callstate_t))) == NULL) {
		debug("bdb_control: asleep: bad alloc");
		ret[0] = 1;
		retlen = 1;		
	    } else {
		c->cmd = command;
		c->invoke = NULL; /* XXX */
		c->free = free_data;
		c->xid = desc->nextxid++;
		if (rlen < 3)
		    abort();	/* XXX */
		ret[0] = 0;	/* XXX */
		ret[1] = (c->xid & 0xff00) >> 8;
		ret[2] = c->xid & 0xff;
		retlen = 3;
		debug("bdb_control: asleep: asleep xid = %d", c->xid);
	    }
	}
	break;
#endif	/* XXX_ASLEEP_EXAMPLE */
    default:
	debug("bdb_control: invalid command %d", command);
	ret[0] = 1;		/* XXX */
	retlen = 1;
	break;
    }
    if (c != NULL) {
	driver_async(desc->port, NULL, c->invoke, c, c->free);
    }
    debug("bdb_control: retlen = %d", retlen);
    return retlen;
}


static void
bdb_outputv(ErlDrvData drv_data, ErlIOVec *ev)
{
    descriptor		*desc = (descriptor *) drv_data;
    unsigned char	cmd;
    int			p = 0, q = 1, i, newp, newq;
    int			tmp;
    callstate_t		*c = NULL;
    char		ch, file[MAXPATHLEN], database[MAXPATHLEN], *pos;
    char		outfile[MAXPATHLEN];
    unsigned char	dbnum = BDB_MAX_TABLES;
    int			txnnum = BDB_MAX_TXNS;
    unsigned char	cursornum = BDB_MAX_CURSORS;

    if (desc == NULL || ev == NULL || ev->size < 1) {
	debug("bdb_outputv: bad arg(s)");
	return;
    }
    if (! EV_GET_CHAR(ev, &cmd, &p, &q)) {
	debug("bdb_outputv: empty command");
	reply_error(desc, EINVAL);
    }
    debug("bdb_outputv: my threadid = %lx, cmd = %d", pthread_self(), cmd);
    switch (cmd) {
    case BDB_DEBUG:
	if (ev->size != 1 + (1 * sizeof(unsigned long))) {
	    debug("%s: bad size", __FUNCTION__);
	    reply_error(desc, EINVAL);
	} else {
	    EV_GET_UINT32(ev, &tmp, &p, &q);
	    reply_ok_num(desc, _debug);	/* Immediate reply */
	    _debug = tmp;
	}
	break;
    case BDB_ENV_CREATE:
	if (desc->dbenvp != NULL) {
	    debug("%s: dbenvp != NULL", __FUNCTION__);
	    reply_error(desc, EINVAL);
	    break;
	}
	if (ev->size != 1 + (1 * sizeof(unsigned long))) {
	    debug("%s: bad size", __FUNCTION__);
	    reply_error(desc, EINVAL);
	} else {
	    if ((c = sys_alloc(sizeof(callstate_t))) == NULL) {
		reply_error(desc, ENOMEM);
	    } else {
		c->cmd = cmd;
		c->key = NULL;
		c->invoke = invoke_db_env_create;
		c->free = free_data;
		EV_GET_UINT32(ev, &c->i.flags, &p, &q);
	    }
	}
	break;
    case BDB_ENV_OPEN:
	if (desc->dbenvp == NULL) {
	    debug("%s: dbenvp == NULL", __FUNCTION__);
	    reply_error(desc, EINVAL);
	    break;
	}
	if ((c = sys_alloc(sizeof(callstate_t))) == NULL) {
	    reply_error(desc, ENOMEM);
	} else {
	    c->cmd = cmd;
	    c->key = NULL;
	    c->invoke = invoke_db_env_open;
	    c->free = free_data;
	    c->i.dbenvp = desc->dbenvp;
	    for (i = 0; i < MAXPATHLEN - 1; i++) {
		EV_GET_CHAR(ev, &ch, &p, &q);
		file[i] = ch;
		if (ch == '\0')
		    break;
	    }
	    c->i.file = file;
	    EV_GET_UINT32(ev, &c->i.flags, &p, &q);
	    EV_GET_UINT32(ev, &c->i.mode, &p, &q);
	}
	break;
    case BDB_ENV_CLOSE:
	if (desc->dbenvp == NULL) {
	    debug("%s: dbenvp == NULL", __FUNCTION__);
	    reply_error(desc, EINVAL);
	    break;
	}
	if ((c = sys_alloc(sizeof(callstate_t))) == NULL) {
	    reply_error(desc, ENOMEM);
	} else {
	    c->cmd = cmd;
	    c->key = NULL;
	    c->invoke = invoke_db_env_close;
	    c->free = free_data;
	    c->i.dbenvp = desc->dbenvp;
	    EV_GET_UINT32(ev, &c->i.flags, &p, &q);
	}
	/* TODO: This isn't safe in the full txn + Pthreads case */
	close_hangers_on(desc);
	break;
    case BDB_ENV_REMOVE:
	if (desc->dbenvp == NULL) {
	    debug("%s: dbenvp == NULL", __FUNCTION__);
	    reply_error(desc, EINVAL);
	    break;
	}
	if ((c = sys_alloc(sizeof(callstate_t))) == NULL) {
	    reply_error(desc, ENOMEM);
	} else {
	    c->cmd = cmd;
	    c->key = NULL;
	    c->invoke = invoke_db_env_remove;
	    c->free = free_data;
	    c->i.dbenvp = desc->dbenvp;
	    EV_GET_CHAR(ev, &dbnum, &p, &q);
	    for (i = 0; i < MAXPATHLEN - 1; i++) {
		EV_GET_CHAR(ev, &ch, &p, &q);
		file[i] = ch;
		if (ch == '\0')
		    break;
	    }
	    c->i.file = file;
	    EV_GET_UINT32(ev, &c->i.flags, &p, &q);
	}
	break;
    case BDB_ENV_SET_RPC_SERVER:
	if (desc->dbenvp == NULL) {
	    debug("%s: dbenvp == NULL", __FUNCTION__);
	    reply_error(desc, EINVAL);
	    break;
	}
	if ((c = sys_alloc(sizeof(callstate_t))) == NULL) {
	    reply_error(desc, ENOMEM);
	} else {
	    c->cmd = cmd;
	    c->key = NULL;
	    c->invoke = invoke_db_env_set_rpc_server;
	    c->free = free_data;
	    c->i.dbenvp = desc->dbenvp;
	    EV_GET_CHAR(ev, &dbnum, &p, &q);
	    for (i = 0; i < MAXPATHLEN - 1; i++) {
		EV_GET_CHAR(ev, &ch, &p, &q);
		file[i] = ch;
		if (ch == '\0')
		    break;
	    }
	    c->i.file = file;
	    EV_GET_UINT32(ev, &c->i.keylen, &p, &q);
	    EV_GET_UINT32(ev, &c->i.vallen, &p, &q);
	    EV_GET_UINT32(ev, &c->i.flags, &p, &q);
	}
	break;
    case BDB_DB_CREATE:
	if (desc->dbenvp == NULL) {
	    debug("%s: dbenvp == NULL", __FUNCTION__);
	    reply_error(desc, EINVAL);
	    break;
	}
	if ((c = sys_alloc(sizeof(callstate_t))) == NULL) {
	    reply_error(desc, ENOMEM);
	} else {
	    c->cmd = cmd;
	    c->key = NULL;
	    c->invoke = invoke_db_create;
	    c->free = free_data;
	    c->i.dbenvp = desc->dbenvp;
	    EV_GET_UINT32(ev, &c->i.flags, &p, &q);
	}
	break;
    case BDB_DB_OPEN:
	if ((c = sys_alloc(sizeof(callstate_t))) == NULL) {
	    reply_error(desc, ENOMEM);
	} else {
	    c->cmd = cmd;
	    c->key = NULL;
	    c->invoke = invoke_db_open;
	    c->free = free_data;
	    c->i.dbenvp = desc->dbenvp;
	    EV_GET_CHAR(ev, &dbnum, &p, &q);
	    for (i = 0; i < MAXPATHLEN - 1; i++) {
		EV_GET_CHAR(ev, &ch, &p, &q);
		file[i] = ch;
		if (ch == '\0')
		    break;
	    }
	    for (i = 0; i < MAXPATHLEN - 1; i++) {
		EV_GET_CHAR(ev, &ch, &p, &q);
		database[i] = ch;
		if (ch == '\0')
		    break;
	    }
	    if (dbnum >= BDB_MAX_TABLES || strlen(file) == 0) {
		sys_free(c);
		c = NULL;
		debug("%s: bad dbnum or file", __FUNCTION__);
		reply_error(desc, EINVAL);
	    } else {
		c->i.dbp = desc->dbp[dbnum];
		c->i.dbnum = dbnum;
		c->i.file = file;
		if (strlen(database) == 0)
		    c->i.database = NULL;
		else
		    c->i.database = database;
		EV_GET_UINT32(ev, &c->i.type, &p, &q);
		EV_GET_UINT32(ev, &c->i.flags, &p, &q);
		EV_GET_UINT32(ev, &c->i.mode, &p, &q);
	    }
	}
	break;
    case BDB_DB_PUT:
	if ((c = sys_alloc(sizeof(callstate_t))) == NULL) {
	    reply_error(desc, ENOMEM);
	} else {
	    c->cmd = cmd;
	    c->key = NULL;
	    c->invoke = invoke_db_put;
	    c->free = free_data;
	    c->i.dbenvp = desc->dbenvp;
	    EV_GET_CHAR(ev, &dbnum, &p, &q);
	    if (dbnum >= BDB_MAX_TABLES  || desc->dbp[dbnum] == NULL) {
		sys_free(c);
		c = NULL;
		debug("%s: bad dbnum or file", __FUNCTION__);
		reply_error(desc, EINVAL);
	    } else {
		c->i.dbp = desc->dbp[dbnum];
		c->i.dbnum = dbnum;
		EV_GET_UINT32(ev, &txnnum, &p, &q);
		if (txnnum < BDB_NULL_TXN || txnnum >= BDB_MAX_TXNS) {
		    debug("XXX QQQ foo 0");
		    abort();	/* XXX */
		}
		if (txnnum >= BDB_MAX_TXNS || desc->txnid[txnnum] == NULL)
		    c->i.txnid = NULL;
		else
		    c->i.txnid = desc->txnid[txnnum];
		EV_GET_UINT32(ev, &c->i.keylen, &p, &q);
		debug("XXX keylen = %d", c->i.keylen);
		pos = EV_GETPOS(ev, p, q);
		newp = p; newq = q;
		if (ev_forward_N(ev, c->i.keylen, &newp, &newq) < 0) {
		    debug("XXX QQQ foo 1");
		    abort();	/* XXX */
		}
		if (newq == q || newq == q + 1) {
		    c->i.key = pos;
		    p = newp; q = newq;
		} else {
		    debug("XXX QQQ foo 2: key spread across multiple iovs");
		    abort();	/* XXX */
		}
		EV_GET_UINT32(ev, &c->i.vallen, &p, &q);
		debug("XXX vallen = %d", c->i.vallen);
		pos = EV_GETPOS(ev, p, q);
		newp = p; newq = q;
		if (ev_forward_N(ev, c->i.vallen, &newp, &newq) < 0) {
		    debug("XXX QQQ foo 3");
		    abort();	/* XXX */
		}
		if (newq == q || newq == q + 1) {
		    c->i.val = pos;
		    p = newp; q = newq;
		} else {
		    debug("XXX QQQ foo 4: val spread across multiple iovs");
		    abort();	/* XXX */
		}
		EV_GET_UINT32(ev, &c->i.flags, &p, &q);
	    }
	}
	break;
    case BDB_DB_GET:
	if ((c = sys_alloc(sizeof(callstate_t))) == NULL) {
	    reply_error(desc, ENOMEM);
	} else {
	    c->cmd = cmd;
	    c->key = NULL;
	    c->invoke = invoke_db_get;
	    c->free = free_data;
	    c->i.dbenvp = desc->dbenvp;
	    EV_GET_CHAR(ev, &dbnum, &p, &q);
	    if (dbnum >= BDB_MAX_TABLES  || desc->dbp[dbnum] == NULL) {
		sys_free(c);
		c = NULL;
		debug("%s: bad dbnum or file", __FUNCTION__);
		reply_error(desc, EINVAL);
	    } else {
		c->i.dbp = desc->dbp[dbnum];
		c->i.dbnum = dbnum;
		EV_GET_UINT32(ev, &txnnum, &p, &q);
		if (txnnum < BDB_NULL_TXN || txnnum >= BDB_MAX_TXNS) {
		    debug("XXX QQQ foo 0");
		    abort();	/* XXX */
		}
		if (txnnum >= BDB_MAX_TXNS || desc->txnid[txnnum] == NULL)
		    c->i.txnid = NULL;
		else
		    c->i.txnid = desc->txnid[txnnum];
		EV_GET_UINT32(ev, &c->i.keylen, &p, &q);
		debug("XXX keylen = %d", c->i.keylen);
		pos = EV_GETPOS(ev, p, q);
		newp = p; newq = q;
		if (ev_forward_N(ev, c->i.keylen, &newp, &newq) < 0) {
		    debug("XXX QQQ foo 1");
		    abort();	/* XXX */
		}
		if (newq == q || newq == q + 1) {
		    c->i.key = pos;
		    p = newp; q = newq;
		} else {
		    debug("XXX QQQ foo 2: key spread across multiple iovs");
		    abort();	/* XXX */
		}
		EV_GET_UINT32(ev, &c->i.flags, &p, &q);
	    }
	}
	break;
    case BDB_DB_DEL:
	if ((c = sys_alloc(sizeof(callstate_t))) == NULL) {
	    reply_error(desc, ENOMEM);
	} else {
	    c->cmd = cmd;
	    c->key = NULL;
	    c->invoke = invoke_db_del;
	    c->free = free_data;
	    c->i.dbenvp = desc->dbenvp;
	    EV_GET_CHAR(ev, &dbnum, &p, &q);
	    if (dbnum >= BDB_MAX_TABLES  || desc->dbp[dbnum] == NULL) {
		sys_free(c);
		c = NULL;
		debug("%s: bad dbnum or file", __FUNCTION__);
		reply_error(desc, EINVAL);
	    } else {
		c->i.dbp = desc->dbp[dbnum];
		c->i.dbnum = dbnum;
		EV_GET_UINT32(ev, &txnnum, &p, &q);
		if (txnnum < BDB_NULL_TXN || txnnum >= BDB_MAX_TXNS) {
		    debug("XXX QQQ foo 0");
		    abort();	/* XXX */
		}
		if (txnnum >= BDB_MAX_TXNS || desc->txnid[txnnum] == NULL)
		    c->i.txnid = NULL;
		else
		    c->i.txnid = desc->txnid[txnnum];
		EV_GET_UINT32(ev, &c->i.keylen, &p, &q);
		debug("XXX keylen = %d", c->i.keylen);
		pos = EV_GETPOS(ev, p, q);
		newp = p; newq = q;
		if (ev_forward_N(ev, c->i.keylen, &newp, &newq) < 0) {
		    debug("XXX QQQ foo 1");
		    abort();	/* XXX */
		}
		if (newq == q || newq == q + 1) {
		    c->i.key = pos;
		    p = newp; q = newq;
		} else {
		    debug("XXX QQQ foo 2: key spread across multiple iovs");
		    abort();	/* XXX */
		}
		EV_GET_UINT32(ev, &c->i.flags, &p, &q);
	    }
	}
	break;
    case BDB_DB_CLOSE:
	if ((c = sys_alloc(sizeof(callstate_t))) == NULL) {
	    reply_error(desc, ENOMEM);
	} else {
	    c->cmd = cmd;
	    c->key = NULL;
	    c->invoke = invoke_db_close;
	    c->free = free_data;
	    c->i.dbenvp = desc->dbenvp;
	    EV_GET_CHAR(ev, &dbnum, &p, &q);
	    for (i = 0; i < BDB_MAX_CURSORS; i++) {
		if (desc->cursorp != NULL && desc->cursor2tab[i] == dbnum) {
		    sys_free(c);
		    c = NULL;
		    debug("%s: table %d has cursor %d open", __FUNCTION__, dbnum, i);
		    reply_error(desc, BDB_ERR_CURSOROPEN);
		    break;
		}
	    }
	    if (dbnum >= BDB_MAX_TABLES || c == NULL) {
		sys_free(c);
		c = NULL;
		debug("%s: bad dbnum or file", __FUNCTION__);
		reply_error(desc, EINVAL);
	    } else {
		c->i.dbp = desc->dbp[dbnum];
		c->i.dbnum = dbnum;
		EV_GET_UINT32(ev, &c->i.flags, &p, &q);
	    }
	}
	break;
    case BDB_DB_UPGRADE:
	if ((c = sys_alloc(sizeof(callstate_t))) == NULL) {
	    reply_error(desc, ENOMEM);
	} else {
	    c->cmd = cmd;
	    c->key = NULL;
	    c->invoke = invoke_db_upgrade;
	    c->free = free_data;
	    c->i.dbenvp = desc->dbenvp;
	    EV_GET_CHAR(ev, &dbnum, &p, &q);
	    if (dbnum >= BDB_MAX_TABLES) {
		sys_free(c);
		c = NULL;
		debug("%s: bad dbnum or file", __FUNCTION__);
		reply_error(desc, EINVAL);
	    } else {
		c->i.dbp = desc->dbp[dbnum];
		c->i.dbnum = dbnum;
		for (i = 0; i < MAXPATHLEN - 1; i++) {
		    EV_GET_CHAR(ev, &ch, &p, &q);
		    file[i] = ch;
		    if (ch == '\0')
			break;
		}
		c->i.file = file;
		EV_GET_UINT32(ev, &c->i.flags, &p, &q);
	    }
	}
	break;
    case BDB_DB_TRUNCATE:
	if ((c = sys_alloc(sizeof(callstate_t))) == NULL) {
	    reply_error(desc, ENOMEM);
	} else {
	    c->cmd = cmd;
	    c->key = NULL;
	    c->invoke = invoke_db_truncate;
	    c->free = free_data;
	    c->i.dbenvp = desc->dbenvp;
	    EV_GET_CHAR(ev, &dbnum, &p, &q);
	    if (dbnum >= BDB_MAX_TABLES  || desc->dbp[dbnum] == NULL) {
		sys_free(c);
		c = NULL;
		debug("%s: bad dbnum or file", __FUNCTION__);
		reply_error(desc, EINVAL);
	    } else {
		c->i.dbp = desc->dbp[dbnum];
		c->i.dbnum = dbnum;
		EV_GET_UINT32(ev, &txnnum, &p, &q);
		if (txnnum < BDB_NULL_TXN || txnnum >= BDB_MAX_TXNS) {
		    debug("XXX QQQ foo 0");
		    abort();	/* XXX */
		}
		if (txnnum >= BDB_MAX_TXNS || desc->txnid[txnnum] == NULL)
		    c->i.txnid = NULL;
		else
		    c->i.txnid = desc->txnid[txnnum];
		EV_GET_UINT32(ev, &c->i.flags, &p, &q);
	    }
	}
	break;
    case BDB_DB_VERIFY:
	if ((c = sys_alloc(sizeof(callstate_t))) == NULL) {
	    reply_error(desc, ENOMEM);
	} else {
	    c->cmd = cmd;
	    c->key = NULL;
	    c->invoke = invoke_db_verify;
	    c->free = free_data;
	    c->i.dbenvp = desc->dbenvp;
	    EV_GET_CHAR(ev, &dbnum, &p, &q);
	    for (i = 0; i < MAXPATHLEN - 1; i++) {
		EV_GET_CHAR(ev, &ch, &p, &q);
		file[i] = ch;
		if (ch == '\0')
		    break;
	    }
	    for (i = 0; i < MAXPATHLEN - 1; i++) {
		EV_GET_CHAR(ev, &ch, &p, &q);
		database[i] = ch;
		if (ch == '\0')
		    break;
	    }
	    for (i = 0; i < MAXPATHLEN - 1; i++) {
		EV_GET_CHAR(ev, &ch, &p, &q);
		outfile[i] = ch;
		if (ch == '\0')
		    break;
	    }
	    if (dbnum >= BDB_MAX_TABLES || strlen(file) == 0) {
		sys_free(c);
		c = NULL;
		debug("%s: bad dbnum or file", __FUNCTION__);
		reply_error(desc, EINVAL);
	    } else {
		c->i.dbp = desc->dbp[dbnum];
		c->i.dbnum = dbnum;
		c->i.file = file;
		if (strlen(database) == 0)
		    c->i.database = NULL;
		else
		    c->i.database = database;
		if (strlen(outfile) == 0)
		    c->i.outfile = NULL;
		else
		    c->i.outfile = outfile;
		EV_GET_UINT32(ev, &c->i.flags, &p, &q);
	    }
	}
	break;
    case BDB_DB_SYNC:
	if ((c = sys_alloc(sizeof(callstate_t))) == NULL) {
	    reply_error(desc, ENOMEM);
	} else {
	    c->cmd = cmd;
	    c->key = NULL;
	    c->invoke = invoke_db_sync;
	    c->free = free_data;
	    c->i.dbenvp = desc->dbenvp;
	    EV_GET_CHAR(ev, &dbnum, &p, &q);
	    if (dbnum >= BDB_MAX_TABLES) {
		sys_free(c);
		c = NULL;
		debug("%s: bad dbnum or file", __FUNCTION__);
		reply_error(desc, EINVAL);
	    } else {
		c->i.dbp = desc->dbp[dbnum];
		c->i.dbnum = dbnum;
		EV_GET_UINT32(ev, &c->i.flags, &p, &q);
	    }
	}
	break;
    case BDB_DB_REMOVE:
	if ((c = sys_alloc(sizeof(callstate_t))) == NULL) {
	    reply_error(desc, ENOMEM);
	} else {
	    c->cmd = cmd;
	    c->key = NULL;
	    c->invoke = invoke_db_remove;
	    c->free = free_data;
	    c->i.dbenvp = desc->dbenvp;
	    EV_GET_CHAR(ev, &dbnum, &p, &q);
	    for (i = 0; i < MAXPATHLEN - 1; i++) {
		EV_GET_CHAR(ev, &ch, &p, &q);
		file[i] = ch;
		if (ch == '\0')
		    break;
	    }
	    for (i = 0; i < MAXPATHLEN - 1; i++) {
		EV_GET_CHAR(ev, &ch, &p, &q);
		database[i] = ch;
		if (ch == '\0')
		    break;
	    }
	    if (dbnum >= BDB_MAX_TABLES || strlen(file) == 0) {
		sys_free(c);
		c = NULL;
		debug("%s: bad dbnum or file", __FUNCTION__);
		reply_error(desc, EINVAL);
	    } else {
		c->i.dbp = desc->dbp[dbnum];
		c->i.dbnum = dbnum;
		c->i.file = file;
		if (strlen(database) == 0)
		    c->i.database = NULL;
		else
		    c->i.database = database;
		EV_GET_UINT32(ev, &c->i.flags, &p, &q);
	    }
	}
	break;
    case BDB_DB_CURSOR:
	if ((c = sys_alloc(sizeof(callstate_t))) == NULL) {
	    reply_error(desc, ENOMEM);
	} else {
	    c->cmd = cmd;
	    c->key = XXXKEY;
	    c->invoke = invoke_db_cursor;
	    c->free = free_data;
	    c->i.dbenvp = desc->dbenvp;
	    EV_GET_CHAR(ev, &dbnum, &p, &q);
	    if (dbnum >= BDB_MAX_TABLES  || desc->dbp[dbnum] == NULL) {
		sys_free(c);
		c = NULL;
		debug("%s: bad dbnum or file", __FUNCTION__);
		reply_error(desc, EINVAL);
	    } else {
		c->i.dbp = desc->dbp[dbnum];
		c->i.dbnum = dbnum;
		EV_GET_UINT32(ev, &txnnum, &p, &q);
		if (txnnum < BDB_NULL_TXN || txnnum >= BDB_MAX_TXNS) {
		    debug("XXX QQQ foo 0");
		    abort();	/* XXX */
		}
		if (txnnum >= BDB_MAX_TXNS || desc->txnid[txnnum] == NULL)
		    c->i.txnid = NULL;
		else
		    c->i.txnid = desc->txnid[txnnum];
		EV_GET_UINT32(ev, &c->i.flags, &p, &q);
	    }
	}
	break;
    case BDB_C_GET:
    case BDB_C_PUT:
	if ((c = sys_alloc(sizeof(callstate_t))) == NULL) {
	    reply_error(desc, ENOMEM);
	} else {
	    c->cmd = cmd;
	    c->key = XXXKEY;
	    if (cmd == BDB_C_PUT) {
		c->invoke = invoke_db_c_put;
	    } else {
		c->invoke = invoke_db_c_get;
	    }
	    c->free = free_data;
	    c->i.dbenvp = desc->dbenvp;
	    EV_GET_CHAR(ev, &cursornum, &p, &q);
	    if (cursornum >= BDB_MAX_CURSORS || desc->cursorp[cursornum] == NULL) {
		sys_free(c);
		c = NULL;
		debug("%s: bad cursornum", __FUNCTION__);
		reply_error(desc, EINVAL);
	    } else {
		c->i.cursorp = desc->cursorp[cursornum];
		EV_GET_UINT32(ev, &c->i.keylen, &p, &q);
		debug("XXX keylen = %d", c->i.keylen);
		pos = EV_GETPOS(ev, p, q);
		newp = p; newq = q;
		if (ev_forward_N(ev, c->i.keylen, &newp, &newq) < 0) {
		    debug("XXX QQQ foo 1");
		    abort();	/* XXX */
		}
		if (newq == q || newq == q + 1) {
		    c->i.key = pos;
		    p = newp; q = newq;
		} else {
		    debug("XXX QQQ foo 2: key spread across multiple iovs");
		    abort();	/* XXX */
		}
		EV_GET_UINT32(ev, &c->i.vallen, &p, &q);
		debug("XXX vallen = %d", c->i.vallen);
		pos = EV_GETPOS(ev, p, q);
		newp = p; newq = q;
		if (ev_forward_N(ev, c->i.vallen, &newp, &newq) < 0) {
		    debug("XXX QQQ foo 3");
		    abort();	/* XXX */
		}
		if (newq == q || newq == q + 1) {
		    c->i.val = pos;
		    p = newp; q = newq;
		} else {
		    debug("XXX QQQ foo 4: val spread across multiple iovs");
		    abort();	/* XXX */
		}
		EV_GET_UINT32(ev, &c->i.flags, &p, &q);
	    }
	}
	break;
    case BDB_C_DEL:
	if ((c = sys_alloc(sizeof(callstate_t))) == NULL) {
	    reply_error(desc, ENOMEM);
	} else {
	    c->cmd = cmd;
	    c->key = XXXKEY;
	    c->invoke = invoke_db_c_del;
	    c->free = free_data;
	    c->i.dbenvp = desc->dbenvp;
	    EV_GET_CHAR(ev, &cursornum, &p, &q);
	    if (cursornum >= BDB_MAX_CURSORS || desc->cursorp[cursornum] == NULL) {
		sys_free(c);
		c = NULL;
		debug("%s: bad cursornum", __FUNCTION__);
		reply_error(desc, EINVAL);
	    } else {
		c->i.cursorp = desc->cursorp[cursornum];
		EV_GET_UINT32(ev, &c->i.flags, &p, &q);
	    }
	}
	break;
    case BDB_C_DUP:
	if ((c = sys_alloc(sizeof(callstate_t))) == NULL) {
	    reply_error(desc, ENOMEM);
	} else {
	    c->cmd = cmd;
	    c->key = XXXKEY;
	    c->invoke = invoke_db_c_dup;
	    c->free = free_data;
	    c->i.dbenvp = desc->dbenvp;
	    EV_GET_CHAR(ev, &cursornum, &p, &q);
	    if (cursornum >= BDB_MAX_CURSORS || desc->cursorp[cursornum] == NULL) {
		sys_free(c);
		c = NULL;
		debug("%s: bad cursornum", __FUNCTION__);
		reply_error(desc, EINVAL);
	    } else {
		c->i.cursorp = desc->cursorp[cursornum];
		EV_GET_UINT32(ev, &c->i.flags, &p, &q);
	    }
	}
	break;
    case BDB_C_COUNT:
	if ((c = sys_alloc(sizeof(callstate_t))) == NULL) {
	    reply_error(desc, ENOMEM);
	} else {
	    c->cmd = cmd;
	    c->key = XXXKEY;
	    c->invoke = invoke_db_c_count;
	    c->free = free_data;
	    c->i.dbenvp = desc->dbenvp;
	    EV_GET_CHAR(ev, &cursornum, &p, &q);
	    if (cursornum >= BDB_MAX_CURSORS || desc->cursorp[cursornum] == NULL) {
		sys_free(c);
		c = NULL;
		debug("%s: bad cursornum", __FUNCTION__);
		reply_error(desc, EINVAL);
	    } else {
		c->i.cursorp = desc->cursorp[cursornum];
		EV_GET_UINT32(ev, &c->i.flags, &p, &q);
	    }
	}
	break;
    case BDB_C_CLOSE:
	if ((c = sys_alloc(sizeof(callstate_t))) == NULL) {
	    reply_error(desc, ENOMEM);
	} else {
	    c->cmd = cmd;
	    c->key = XXXKEY;
	    c->invoke = invoke_db_c_close;
	    c->free = free_data;
	    c->i.dbenvp = desc->dbenvp;
	    EV_GET_CHAR(ev, &cursornum, &p, &q);
	    if (cursornum >= BDB_MAX_CURSORS || desc->cursorp[cursornum] == NULL) {
		sys_free(c);
		c = NULL;
		debug("%s: bad cursornum", __FUNCTION__);
		reply_error(desc, EINVAL);
	    } else {
		c->i.cursorp = desc->cursorp[cursornum];
		c->i.cursornum = cursornum;
	    }
	}
	break;
    case BDB_DB_SET_PAGESIZE:
	if ((c = sys_alloc(sizeof(callstate_t))) == NULL) {
	    reply_error(desc, ENOMEM);
	} else {
	    c->cmd = cmd;
	    c->key = NULL;
	    c->invoke = invoke_db_set_pagesize;
	    c->free = free_data;
	    c->i.dbenvp = desc->dbenvp;
	    EV_GET_CHAR(ev, &dbnum, &p, &q);
	    if (dbnum >= BDB_MAX_TABLES) {
		sys_free(c);
		c = NULL;
		debug("%s: bad dbnum or file", __FUNCTION__);
		reply_error(desc, EINVAL);
	    } else {
		c->i.dbp = desc->dbp[dbnum];
		c->i.dbnum = dbnum;
		EV_GET_UINT32(ev, &c->i.flags, &p, &q);
	    }
	}
	break;
    case BDB_DB_SET_CACHESIZE:
	if ((c = sys_alloc(sizeof(callstate_t))) == NULL) {
	    reply_error(desc, ENOMEM);
	} else {
	    c->cmd = cmd;
	    c->key = NULL;
	    c->invoke = invoke_db_set_cachesize;
	    c->free = free_data;
	    c->i.dbenvp = desc->dbenvp;
	    EV_GET_CHAR(ev, &dbnum, &p, &q);
	    if (dbnum >= BDB_MAX_TABLES) {
		sys_free(c);
		c = NULL;
		debug("%s: bad dbnum or file", __FUNCTION__);
		reply_error(desc, EINVAL);
	    } else {
		c->i.dbp = desc->dbp[dbnum];
		c->i.dbnum = dbnum;
		EV_GET_UINT32(ev, &c->i.keylen, &p, &q);
		EV_GET_UINT32(ev, &c->i.vallen, &p, &q);
		EV_GET_UINT32(ev, &c->i.mode, &p, &q);
	    }
	}
	break;
    case BDB_DB_SET_LORDER:
	if ((c = sys_alloc(sizeof(callstate_t))) == NULL) {
	    reply_error(desc, ENOMEM);
	} else {
	    c->cmd = cmd;
	    c->key = NULL;
	    c->invoke = invoke_db_set_lorder;
	    c->free = free_data;
	    c->i.dbenvp = desc->dbenvp;
	    EV_GET_CHAR(ev, &dbnum, &p, &q);
	    if (dbnum >= BDB_MAX_TABLES) {
		sys_free(c);
		c = NULL;
		debug("%s: bad dbnum or file", __FUNCTION__);
		reply_error(desc, EINVAL);
	    } else {
		c->i.dbp = desc->dbp[dbnum];
		c->i.dbnum = dbnum;
		EV_GET_UINT32(ev, &c->i.mode, &p, &q);
	    }
	}
	break;
    case BDB_DB_SET_FLAGS:
	if ((c = sys_alloc(sizeof(callstate_t))) == NULL) {
	    reply_error(desc, ENOMEM);
	} else {
	    c->cmd = cmd;
	    c->key = NULL;
	    c->invoke = invoke_db_set_flags;
	    c->free = free_data;
	    c->i.dbenvp = desc->dbenvp;
	    EV_GET_CHAR(ev, &dbnum, &p, &q);
	    if (dbnum >= BDB_MAX_TABLES) {
		sys_free(c);
		c = NULL;
		debug("%s: bad dbnum or file", __FUNCTION__);
		reply_error(desc, EINVAL);
	    } else {
		c->i.dbp = desc->dbp[dbnum];
		c->i.dbnum = dbnum;
		EV_GET_UINT32(ev, &c->i.flags, &p, &q);
	    }
	}
	break;
    case BDB_DB_SET_BT_MINKEY:
	if ((c = sys_alloc(sizeof(callstate_t))) == NULL) {
	    reply_error(desc, ENOMEM);
	} else {
	    c->cmd = cmd;
	    c->key = NULL;
	    c->invoke = invoke_db_set_bt_minkey;
	    c->free = free_data;
	    c->i.dbenvp = desc->dbenvp;
	    EV_GET_CHAR(ev, &dbnum, &p, &q);
	    if (dbnum >= BDB_MAX_TABLES) {
		sys_free(c);
		c = NULL;
		debug("%s: bad dbnum or file", __FUNCTION__);
		reply_error(desc, EINVAL);
	    } else {
		c->i.dbp = desc->dbp[dbnum];
		c->i.dbnum = dbnum;
		EV_GET_UINT32(ev, &c->i.flags, &p, &q);
	    }
	}
	break;
    case BDB_DB_SET_H_FFACTOR:
	if ((c = sys_alloc(sizeof(callstate_t))) == NULL) {
	    reply_error(desc, ENOMEM);
	} else {
	    c->cmd = cmd;
	    c->key = NULL;
	    c->invoke = invoke_db_set_h_ffactor;
	    c->free = free_data;
	    c->i.dbenvp = desc->dbenvp;
	    EV_GET_CHAR(ev, &dbnum, &p, &q);
	    if (dbnum >= BDB_MAX_TABLES) {
		sys_free(c);
		c = NULL;
		debug("%s: bad dbnum or file", __FUNCTION__);
		reply_error(desc, EINVAL);
	    } else {
		c->i.dbp = desc->dbp[dbnum];
		c->i.dbnum = dbnum;
		EV_GET_UINT32(ev, &c->i.flags, &p, &q);
	    }
	}
	break;
    case BDB_DB_SET_H_NELEM:
	if ((c = sys_alloc(sizeof(callstate_t))) == NULL) {
	    reply_error(desc, ENOMEM);
	} else {
	    c->cmd = cmd;
	    c->key = NULL;
	    c->invoke = invoke_db_set_h_nelem;
	    c->free = free_data;
	    c->i.dbenvp = desc->dbenvp;
	    EV_GET_CHAR(ev, &dbnum, &p, &q);
	    if (dbnum >= BDB_MAX_TABLES) {
		sys_free(c);
		c = NULL;
		debug("%s: bad dbnum or file", __FUNCTION__);
		reply_error(desc, EINVAL);
	    } else {
		c->i.dbp = desc->dbp[dbnum];
		c->i.dbnum = dbnum;
		EV_GET_UINT32(ev, &c->i.flags, &p, &q);
	    }
	}
	break;
    case BDB_DB_SET_RE_DELIM:
	if ((c = sys_alloc(sizeof(callstate_t))) == NULL) {
	    reply_error(desc, ENOMEM);
	} else {
	    c->cmd = cmd;
	    c->key = NULL;
	    c->invoke = invoke_db_set_re_delim;
	    c->free = free_data;
	    c->i.dbenvp = desc->dbenvp;
	    EV_GET_CHAR(ev, &dbnum, &p, &q);
	    if (dbnum >= BDB_MAX_TABLES) {
		sys_free(c);
		c = NULL;
		debug("%s: bad dbnum or file", __FUNCTION__);
		reply_error(desc, EINVAL);
	    } else {
		c->i.dbp = desc->dbp[dbnum];
		c->i.dbnum = dbnum;
		EV_GET_UINT32(ev, &c->i.mode, &p, &q);
	    }
	}
	break;
    case BDB_DB_SET_RE_LEN:
	if ((c = sys_alloc(sizeof(callstate_t))) == NULL) {
	    reply_error(desc, ENOMEM);
	} else {
	    c->cmd = cmd;
	    c->key = NULL;
	    c->invoke = invoke_db_set_re_len;
	    c->free = free_data;
	    c->i.dbenvp = desc->dbenvp;
	    EV_GET_CHAR(ev, &dbnum, &p, &q);
	    if (dbnum >= BDB_MAX_TABLES) {
		sys_free(c);
		c = NULL;
		debug("%s: bad dbnum or file", __FUNCTION__);
		reply_error(desc, EINVAL);
	    } else {
		c->i.dbp = desc->dbp[dbnum];
		c->i.dbnum = dbnum;
		EV_GET_UINT32(ev, &c->i.flags, &p, &q);
	    }
	}
	break;
    case BDB_DB_SET_RE_PAD:
	if ((c = sys_alloc(sizeof(callstate_t))) == NULL) {
	    reply_error(desc, ENOMEM);
	} else {
	    c->cmd = cmd;
	    c->key = NULL;
	    c->invoke = invoke_db_set_re_pad;
	    c->free = free_data;
	    c->i.dbenvp = desc->dbenvp;
	    EV_GET_CHAR(ev, &dbnum, &p, &q);
	    if (dbnum >= BDB_MAX_TABLES) {
		sys_free(c);
		c = NULL;
		debug("%s: bad dbnum or file", __FUNCTION__);
		reply_error(desc, EINVAL);
	    } else {
		c->i.dbp = desc->dbp[dbnum];
		c->i.dbnum = dbnum;
		EV_GET_UINT32(ev, &c->i.mode, &p, &q);
	    }
	}
	break;
    case BDB_DB_SET_RE_SOURCE:
	if ((c = sys_alloc(sizeof(callstate_t))) == NULL) {
	    reply_error(desc, ENOMEM);
	} else {
	    c->cmd = cmd;
	    c->key = NULL;
	    c->invoke = invoke_db_set_re_source;
	    c->free = free_data;
	    c->i.dbenvp = desc->dbenvp;
	    EV_GET_CHAR(ev, &dbnum, &p, &q);
	    if (dbnum >= BDB_MAX_TABLES) {
		sys_free(c);
		c = NULL;
		debug("%s: bad dbnum or file", __FUNCTION__);
		reply_error(desc, EINVAL);
	    } else {
		c->i.dbp = desc->dbp[dbnum];
		c->i.dbnum = dbnum;
		EV_GET_CHAR(ev, &dbnum, &p, &q);
		for (i = 0; i < MAXPATHLEN - 1; i++) {
		    EV_GET_CHAR(ev, &ch, &p, &q);
		    file[i] = ch;
		    if (ch == '\0')
			break;
		}
		c->i.file = file;
	    }
	}
#ifdef	XXX_TODO
    case BDB_TXN_BEGIN:
	if ((c = sys_alloc(sizeof(callstate_t))) == NULL) {
	    reply_error(desc, ENOMEM);
	} else {
	    c->cmd = cmd;
	    c->key = NULL;
	    c->invoke = invoke_db_txn_begin;
	    c->free = free_data;
	    c->i.dbenvp = desc->dbenvp;
	    EV_GET_CHAR(ev, &dbnum, &p, &q);
	    if (dbnum >= BDB_MAX_TABLES  || desc->dbp[dbnum] == NULL) {
		sys_free(c);
		c = NULL;
		debug("%s: bad dbnum or file", __FUNCTION__);
		reply_error(desc, EINVAL);
	    } else {
		c->i.dbp = desc->dbp[dbnum];
		c->i.dbnum = dbnum;
		EV_GET_UINT32(ev, &txnnum, &p, &q);
		if (txnnum < BDB_NULL_TXN || txnnum >= BDB_MAX_TXNS) {
		    debug("XXX QQQ foo 0");
		    abort();	/* XXX */
		}
		if (txnnum >= BDB_MAX_TXNS || desc->txnid[txnnum] == NULL)
		    c->i.txnid = NULL;
		else
		    c->i.txnid = desc->txnid[txnnum];
		EV_GET_UINT32(ev, &c->i.flags, &p, &q);
	    }
	}
	break;
#endif	/* XXX_TODO */
    default:
	debug("bdb_outputv: invalid command %d", cmd);
	break;
    }
    if (c != NULL) {
	if (c->key == NULL) {
	    driver_async(desc->port, c->key, c->invoke, c, c->free);
	} else {
	    (*(c->invoke))((void *) c);
	    bdb_ready_async((ErlDrvData) desc, (ErlDrvThreadData) c);
	    /* c is already freed for us by bdb_ready_async() */
	}
    }
}

static void
bdb_ready_async(ErlDrvData drv_data, ErlDrvThreadData thread_data)
{
    descriptor	*desc = (descriptor *) drv_data;
    callstate_t	*c = (callstate_t *) thread_data;
    int		i;
    DB		*dbp;
    DBC		*cursorp;

    debug("bdb_async_ready: cmd = %d status = %d", c->cmd, c->status);
    if (c == NULL) {
	debug("%s: c == NULL", __FUNCTION__);
	return;
    }
    switch (c->cmd) {
    case BDB_ENV_CREATE:
	if (c->status == 0) {
	    desc->dbenvp = c->o.ptr;
	    desc->dbenvp->set_errcall(desc->dbenvp, debug_errcall);
	    if ((i = desc->dbenvp->set_alloc(desc->dbenvp,
					     driver_alloc_wrapper, 
					     driver_realloc_wrapper,
					     driver_free_wrapper)) != 0) {
		reply_error(desc, i);
	    } else {
		debug("XXX %s: DB_ENV->set_alloc() claimed success!!!", __FUNCTION__);
		reply_ok(desc);
	    }
	} else {
	    reply_error(desc, c->status);
	}
	break;
    case BDB_ENV_CLOSE:
	if (c->status == 0) {
	    desc->dbenvp = NULL;
	    reply_ok(desc);
	} else {
	    reply_error(desc, c->status);
	}
	break;
    case BDB_DB_CREATE:
	if (c->status == 0) {
	    dbp = (DB *) c->o.ptr;
	    for (i = 0; i < BDB_MAX_TABLES; i++) {
		if (desc->dbp[i] == NULL)
		    break;
	    }
	    if (i == BDB_MAX_TABLES) {
		dbp->close(dbp, 0);
		reply_error(desc, EMFILE);
	    } else {
		desc->dbp[i] = dbp;
		reply_ok_num(desc, i);
	    }
	} else {
	    reply_error(desc, c->status);
	}
	break;
    case BDB_DB_CURSOR:
    case BDB_C_DUP:
	if (c->status == 0) {
	    cursorp = (DBC *) c->o.ptr;
	    for (i = 0; i < BDB_MAX_CURSORS; i++) {
		if (desc->cursorp[i] == NULL)
		    break;
	    }
	    if (i == BDB_MAX_CURSORS) {
		cursorp->c_close(cursorp);
		reply_error(desc, EMFILE);
	    } else {
		desc->cursorp[i] = cursorp;
		desc->cursor2tab[i] = c->i.dbnum;
		reply_ok_num(desc, i);
	    }
	} else {
	    reply_error(desc, c->status);
	}
	break;
    case BDB_ENV_OPEN:
    case BDB_ENV_REMOVE:
    case BDB_ENV_SET_RPC_SERVER:
    case BDB_DB_OPEN:
    case BDB_DB_PUT:
    case BDB_DB_DEL:
    case BDB_DB_UPGRADE:
    case BDB_DB_VERIFY:
    case BDB_DB_SYNC:
    case BDB_DB_REMOVE:
    case BDB_C_PUT:
    case BDB_C_DEL:
    case BDB_DB_SET_PAGESIZE:
    case BDB_DB_SET_CACHESIZE:
    case BDB_DB_SET_LORDER:
    case BDB_DB_SET_FLAGS:
    case BDB_DB_SET_BT_MINKEY:
    case BDB_DB_SET_H_FFACTOR:
    case BDB_DB_SET_H_NELEM:
    case BDB_DB_SET_RE_DELIM:
    case BDB_DB_SET_RE_LEN:
    case BDB_DB_SET_RE_PAD:
    case BDB_DB_SET_RE_SOURCE:
	if (c->status == 0) {
	    reply_ok(desc);
	} else {
	    reply_error(desc, c->status);
	}
	break;
    case BDB_DB_GET:
    case BDB_C_GET:
	if (c->status == 0) {
	    reply_ok_key_val(desc, c);
	} else {
	    reply_error(desc, c->status);
	}
	break;
    case BDB_DB_CLOSE:
	if (c->status == 0) {
	    desc->dbp[(int) c->i.dbnum] = NULL;
	    reply_ok(desc);
	} else {
	    reply_error(desc, c->status);
	}
	break;
    case BDB_C_CLOSE:
	if (c->status == 0) {
	    desc->cursorp[(int) c->i.cursornum] = NULL;
	    desc->cursor2tab[(int) c->i.cursornum] = -1;
	    reply_ok(desc);
	} else {
	    reply_error(desc, c->status);
	}
	break;
    case BDB_DB_TRUNCATE:
	if (c->status == 0) {
	    reply_ok_num(desc, c->o.vallen);
	} else {
	    reply_error(desc, c->status);
	}
	break;
    case BDB_C_COUNT:
	if (c->status == 0) {
	    reply_ok_num(desc, c->o.vallen);
	} else {
	    reply_error(desc, c->status);
	}
	break;
    case -666 /* XXX */:
	if (c->status == 0) {
	    reply_tag_ok(desc, c->xid);
	} else {
	    reply_tag_error(desc, c->xid, c->status);
	}
	break;
    default:
	debug("bdb_async_ready: bogus command, should never happen");
	break;
    }    
    sys_free(c);
}

static void *
driver_alloc_wrapper(size_t size)
{
    ErlDrvBinary	*eb;

    debug("%s: top", __FUNCTION__);
    if ((eb = driver_alloc_binary(size)) != NULL) {
	debug("%s: size %lu eb = 0x%lx orig_bytes = 0x%lx", __FUNCTION__, size, eb, eb->orig_bytes);
	return eb->orig_bytes;
    } else {
	return NULL;
    }
}

static void *
driver_realloc_wrapper(void *p, size_t size)
{
    ErlDrvBinary	*eb, *neweb;

    debug("%s: top", __FUNCTION__);
    if ((eb = alloced_ptr2ErlDrvBinary(p)) != NULL) {
	if ((neweb = driver_realloc_binary(eb, size)) != NULL) {
	    return neweb->orig_bytes;
	} else {
	    return NULL;
	}
    } else {
	return NULL;
    }
}

static void
driver_free_wrapper(void *p)
{
    ErlDrvBinary	*eb;

    debug("%s: top", __FUNCTION__);
    if ((eb = alloced_ptr2ErlDrvBinary(p)) != NULL) {
	driver_free_binary(eb);
    }
}

/* XXX Make this a macro, use scratch static ErlDrvBinary? */
static ErlDrvBinary *
alloced_ptr2ErlDrvBinary(void *p)
{
    ErlDrvBinary	stackeb;
    int			offset = stackeb.orig_bytes - (char *) &stackeb;

    if (p != NULL) {
	debug("%s: p = 0x%lx eb = 0x%lx", __FUNCTION__, p, (ErlDrvBinary *) ((char *) p - offset));
	return (ErlDrvBinary *) ((char *) p - offset);
    } else {
	return NULL;
    }
}

static int
reply_ok(descriptor *desc)
{
    ErlDrvTermData	msg[15]; /* XXX too big */
    int			i = 0;
    int			res;

    i = LOAD_PORT(msg, i, driver_mk_port(desc->port));
    i = LOAD_ATOM(msg, i, am_ok);
    i = LOAD_TUPLE(msg, i, 2);
    debug("reply_ok: i = %d", i);
    res = driver_output_term(desc->port, msg, i);
    debug("reply_ok: res = %d", res);
    return res;
}

static int
reply_ok_num(descriptor *desc, u_int32_t num)
{
    ErlDrvTermData	msg[15]; /* XXX too big */
    int			i = 0;
    int			res;

    i = LOAD_PORT(msg, i, driver_mk_port(desc->port));
    i = LOAD_ATOM(msg, i, am_ok);
    i = LOAD_INT(msg, i, num);
    i = LOAD_TUPLE(msg, i, 3);
    debug("%s: i = %d, num = %lu", __FUNCTION__, i, num);
    res = driver_output_term(desc->port, msg, i);
    debug("%s: res = %d", __FUNCTION__, res);
    return res;
}

static int
reply_ok_key_val(descriptor *desc, callstate_t *c)
{
    ErlDrvTermData	msg[15]; /* XXX too big */
    int			i = 0;
    int			res;

    debug("%s: result keylen = %d result vallen = %d", __FUNCTION__, c->o.keylen, c->o.vallen);
    i = LOAD_PORT(msg, i, driver_mk_port(desc->port));
    i = LOAD_ATOM(msg, i, am_ok);
    {
	ErlDrvBinary *e = alloced_ptr2ErlDrvBinary(c->o.key);
	debug("\t%s: XXX: key: orig_size = %d, refc = %d, keylen = %d", __FUNCTION__, e->orig_size, e->refc, c->o.keylen);
    }
    {
	ErlDrvBinary *e = alloced_ptr2ErlDrvBinary(c->o.val);
	debug("\t%s: XXX: val: orig_size = %d, refc = %d, vallen = %d", __FUNCTION__, e->orig_size, e->refc, c->o.vallen);
    }
    i = LOAD_BINARY(msg, i, alloced_ptr2ErlDrvBinary(c->o.key), 0, c->o.keylen);
    i = LOAD_BINARY(msg, i, alloced_ptr2ErlDrvBinary(c->o.val), 0, c->o.vallen);
    i = LOAD_TUPLE(msg, i, 4);
    /* QQQ debug("%s: i = %d", __FUNCTION__, i); */
    /* QQQ debug("XXX YYY ZZZ\tkey refc before = %d", alloced_ptr2ErlDrvBinary(c->o.key)->refc); */
    res = driver_output_term(desc->port, msg, i);
    /* QQQ debug("XXX YYY ZZZ\tkey refc after output = %d", alloced_ptr2ErlDrvBinary(c->o.key)->refc); */
    driver_free_binary(alloced_ptr2ErlDrvBinary(c->o.key));
    driver_free_binary(alloced_ptr2ErlDrvBinary(c->o.val));
    /* QQQ debug("XXX YYY ZZZ\tkey refc after driver_free = %d", alloced_ptr2ErlDrvBinary(c->o.key)->refc); */
    /* QQQ debug("%s: res = %d", __FUNCTION__, res); */
    return res;
}

static int
reply_error(descriptor *desc, int errnum)
{
    ErlDrvTermData	msg[15]; /* XXX too big */
    int			i = 0;
    int			res;

    debug("reply_error: errnum = %d", errnum);
    i = LOAD_PORT(msg, i, driver_mk_port(desc->port));
    i = LOAD_ATOM(msg, i, am_error);
    i = LOAD_INT(msg, i, errnum);
    i = LOAD_TUPLE(msg, i, 3);
    debug("reply_error: i = %d", i);
    res = driver_output_term(desc->port, msg, i);
    debug("reply_error: res = %d", res);
    return res;
}

static int
reply_tag_ok(descriptor *desc, unsigned short tag)
{
    ErlDrvTermData	msg[15]; /* XXX too big */
    int			i = 0;
    int			res;

    i = LOAD_PORT(msg, i, driver_mk_port(desc->port));
    i = LOAD_INT(msg, i, tag);
    i = LOAD_ATOM(msg, i, am_ok);
    i = LOAD_TUPLE(msg, i, 3);
    debug("reply_ok: i = %d", i);
    res = driver_output_term(desc->port, msg, i);
    debug("reply_ok: res = %d", res);
    return res;
}

static int
reply_tag_error(descriptor *desc, unsigned short tag, int errnum)
{
    ErlDrvTermData	msg[15]; /* XXX too big */
    int			i = 0;
    int			res;

    debug("reply_error: errnum = %d", errnum);
    i = LOAD_PORT(msg, i, driver_mk_port(desc->port));
    i = LOAD_INT(msg, i, tag);
    i = LOAD_INT(msg, i, errnum);
    i = LOAD_ATOM(msg, i, am_error);
    i = LOAD_TUPLE(msg, i, 4);
    debug("reply_error: i = %d", i);
    res = driver_output_term(desc->port, msg, i);
    debug("reply_error: res = %d", res);
    return res;
}

static int
debug(char *f, ...)
{
    va_list	ap;
    int		res;

    va_start(ap, f);
    if (_debug) {
	res = vfprintf(DEBUGIO, f, ap);
	fprintf(DEBUGIO, "\r\n");
    }
    va_end(ap);
    return res;
}

static void 
debug_errcall(const char *errpfx, char *msg)
{
    if (_debug)
	fprintf(stderr, "\t\t\t\t\t%s: %s: %s\r\n", __FUNCTION__, errpfx, msg);
}

static void
free_data(void *data)
{
    if (data != NULL)
	sys_free(data);
}

/*
** Return value:
**      -1 = error
**       0 = Success, there is no more data to be read
**       1 = Success
*/

static int
ev_forward_N(ErlIOVec *ev, int n, int *pp, int *qp)
{
    int	forward_pos = *pp + n;

    if ((*qp) >= ev->vsize) {
	return -1;
    }
    if (forward_pos < ev->iov[*qp].iov_len) {
	*pp += n;
	return 1;
    } else if (forward_pos == ev->iov[*qp].iov_len) {
	(*qp)++;
	*pp = 0;
	if ((*qp) < ev->vsize) {
	    return 1;
	} else {
	    return 0;
	}
    } else {
	n -= (ev->iov[*qp].iov_len - *pp);
	(*qp)++;
	(*pp) = 0;
	if ((*qp) < ev->vsize) {
	    return ev_forward_N(ev, n, pp, qp);
	} else {
	    return -1;
	}
    }
}

static void
invoke_db_env_create(void *data)
{
    callstate_t	*c = (callstate_t *) data;

    debug("%s: threadid = %lx", __FUNCTION__, pthread_self());
    if ((c->status = db_env_create((DB_ENV **) &c->o.ptr, c->i.flags)) != 0) {
	debug("%s: errno = %d", __FUNCTION__, c->status);
    }
    debug("%s: threadid = %lx done", __FUNCTION__, pthread_self());
}

static void
invoke_db_env_open(void *data)
{
    callstate_t	*c = (callstate_t *) data;
    u_int32_t	flags = c->i.flags | DB_THREAD;	/* Always use DB_THREAD */

    debug("%s: threadid = %lx", __FUNCTION__, pthread_self());
#ifdef	XXX_TODO
    /*
    ** If the caller asks to initialize the transaction environment, then
    ** we *must* deal with deadlock detection.  I don't know enough of the
    ** subtleties in how often one should check ... so we'll use the
    ** automatic checking mechanism for now.  If the overhead sucks, we
    ** could remove this call to set_lk_detect and require the Erlang
    ** caller to call it before opening the environment.  For now, though,
    ** we'll protect the innocent from themselves....
    **
    ** We could perhaps put the set_lk_detect call in the startup() call
    ** up in Erlang Land, but we'll do it here for now....
    */
    if (flags & DB_INIT_TXN &&
	(c->status = c->i.dbenvp->set_lk_detect(c->i.dbenvp, DB_LOCK_DEFAULT)) != 0) {
	debug("%s: set_lk_detect: errno = %d", __FUNCTION__, c->status);
	return;
    }
#endif	/* XXX_TODO */
    if ((c->status = c->i.dbenvp->open(c->i.dbenvp, c->i.file, flags,
				       c->i.mode)) != 0) {
	debug("%s: errno = %d", __FUNCTION__, c->status);
    }
    debug("%s: threadid = %lx done", __FUNCTION__, pthread_self());
}

static void
invoke_db_env_close(void *data)
{
    callstate_t	*c = (callstate_t *) data;

    debug("%s: threadid = %lx", __FUNCTION__, pthread_self());
    if ((c->status = c->i.dbenvp->close(c->i.dbenvp, c->i.flags)) != 0) {
	debug("%s: errno = %d", __FUNCTION__, c->status);
    }
    debug("%s: threadid = %lx done", __FUNCTION__, pthread_self());
}

static void
invoke_db_env_remove(void *data)
{
    callstate_t	*c = (callstate_t *) data;

    debug("%s: threadid = %lx", __FUNCTION__, pthread_self());
    if ((c->status = c->i.dbenvp->remove(c->i.dbenvp, c->i.file,
					 c->i.flags)) != 0) {
	debug("%s: errno = %d", __FUNCTION__, c->status);
    }
    debug("%s: threadid = %lx done", __FUNCTION__, pthread_self());
}

static void
invoke_db_env_set_rpc_server(void *data)
{
    callstate_t	*c = (callstate_t *) data;

    debug("%s: threadid = %lx", __FUNCTION__, pthread_self());
    if ((c->status = c->i.dbenvp->set_rpc_server(c->i.dbenvp, NULL, c->i.file,
						 c->i.keylen, c->i.vallen,
						 c->i.flags)) != 0) {
	debug("%s: errno = %d", __FUNCTION__, c->status);
    }
    debug("%s: threadid = %lx done", __FUNCTION__, pthread_self());
}

static void
invoke_db_create(void *data)
{
    callstate_t	*c = (callstate_t *) data;

    debug("%s: threadid = %lx", __FUNCTION__, pthread_self());
    if ((c->status = db_create((DB **) &c->o.ptr, c->i.dbenvp, c->i.flags)) != 0) {
	debug("%s: errno = %d", __FUNCTION__, c->status);
    }
    debug("%s: threadid = %lx done", __FUNCTION__, pthread_self());
}

static void
invoke_db_open(void *data)
{
    callstate_t	*c = (callstate_t *) data;
    DB		*dbp = c->i.dbp;
    u_int32_t	flags = c->i.flags | DB_THREAD;	/* Always use DB_THREAD */
    debug("%s: threadid = %lx", __FUNCTION__, pthread_self());
    if ((c->status = dbp->open(dbp, c->i.file, c->i.database, c->i.type, 
			       flags, c->i.mode)) != 0) {
	debug("%s: errno = %d", __FUNCTION__, c->status);
	debug("%s: dbp = 0x%lx, file = %s, database = %s, type = %d, flags = %d, mode = %o", __FUNCTION__, dbp, c->i.file, c->i.database, c->i.type, flags, c->i.mode);
    }
    debug("%s: threadid = %lx done", __FUNCTION__, pthread_self());
}

static void
invoke_db_put(void *data)
{
    callstate_t	*c = (callstate_t *) data;
    DB		*dbp = c->i.dbp;
    DBT		k, d;

    debug("%s: threadid = %lx", __FUNCTION__, pthread_self());
    memset(&k, 0, sizeof(k));
    memset(&d, 0, sizeof(d));
    k.data = c->i.key;
    k.size = c->i.keylen;
    d.data = c->i.val;
    d.size = c->i.vallen;
    if ((c->status = dbp->put(dbp, c->i.txnid, &k, &d, c->i.flags)) != 0) {
	debug("%s: errno = %d", __FUNCTION__, c->status);
    }
    debug("%s: threadid = %lx done", __FUNCTION__, pthread_self());
}

static void
invoke_db_get(void *data)
{
    callstate_t		*c = (callstate_t *) data;
    DB			*dbp = c->i.dbp;
    DBT			k, d;

    debug("%s: threadid = %lx", __FUNCTION__, pthread_self());
    memset(&k, 0, sizeof(k));
    memset(&d, 0, sizeof(d));
    k.data = c->i.key;
    k.size = c->i.keylen;
    k.flags = DB_DBT_MALLOC;
    d.size = c->i.vallen;
    d.flags = DB_DBT_MALLOC;
    if ((c->status = dbp->get(dbp, c->i.txnid, &k, &d, c->i.flags)) != 0) {
	debug("%s: errno = %d", __FUNCTION__, c->status);
    } else {
	if (k.data == c->i.key) {
	    if ((c->o.key = driver_alloc_wrapper(k.size)) == NULL) {
		debug("%s: driver_alloc_wrapper failed", __FUNCTION__);
	    }
	    memcpy(c->o.key, k.data, k.size);
	} else {
	    /* XXX Is this ever the case? */
	    c->o.key = k.data;
	}
	c->o.keylen = k.size;
	c->o.val = d.data;
	c->o.vallen = d.size;
    }
    debug("%s: threadid = %lx done", __FUNCTION__, pthread_self());
}

static void
invoke_db_del(void *data)
{
    callstate_t	*c = (callstate_t *) data;
    DB		*dbp = c->i.dbp;
    DBT		k;

    debug("%s: threadid = %lx", __FUNCTION__, pthread_self());
    memset(&k, 0, sizeof(k));
    k.data = c->i.key;
    k.size = c->i.keylen;
    if ((c->status = dbp->del(dbp, c->i.txnid, &k, c->i.flags)) != 0) {
	debug("%s: errno = %d", __FUNCTION__, c->status);
    }
    debug("%s: threadid = %lx done", __FUNCTION__, pthread_self());
}

static void
invoke_db_close(void *data)
{
    callstate_t	*c = (callstate_t *) data;
    DB		*dbp = c->i.dbp;

    debug("%s: threadid = %lx", __FUNCTION__, pthread_self());
    if ((c->status = dbp->close(dbp, c->i.flags)) != 0) {
	debug("%s: errno = %d", __FUNCTION__, c->status);
    }
    debug("%s: threadid = %lx done", __FUNCTION__, pthread_self());
}

static void
invoke_db_upgrade(void *data)
{
    callstate_t	*c = (callstate_t *) data;
    DB		*dbp = c->i.dbp;

    debug("%s: threadid = %lx", __FUNCTION__, pthread_self());
    if ((c->status = dbp->upgrade(dbp, c->i.file, c->i.flags)) != 0) {
	debug("%s: errno = %d", __FUNCTION__, c->status);
    }
    debug("%s: threadid = %lx done", __FUNCTION__, pthread_self());
}

static void
invoke_db_truncate(void *data)
{
    callstate_t	*c = (callstate_t *) data;
    DB		*dbp = c->i.dbp;

    debug("%s: threadid = %lx", __FUNCTION__, pthread_self());
    if ((c->status = dbp->truncate(dbp, c->i.txnid, &c->o.vallen, c->i.flags)) != 0) {
	debug("%s: errno = %d", __FUNCTION__, c->status);
    }
    debug("%s: threadid = %lx done", __FUNCTION__, pthread_self());
}

static void
invoke_db_verify(void *data)
{
    callstate_t	*c = (callstate_t *) data;
    DB		*dbp = c->i.dbp;
    FILE	*fp = NULL;

    debug("%s: threadid = %lx", __FUNCTION__, pthread_self());
    if (c->i.outfile != NULL) {
	/* XXX: Give user ability to pass fopen's 2nd arg? */
	fp = fopen(c->i.outfile, "w");
    }
    if ((c->status = dbp->verify(dbp, c->i.file, c->i.database,
				 fp, c->i.flags)) != 0) {
	debug("%s: errno = %d", __FUNCTION__, c->status);
    }
    if (fp != NULL) {
	fclose(fp);
    }
    debug("%s: threadid = %lx done", __FUNCTION__, pthread_self());
}

static void
invoke_db_sync(void *data)
{
    callstate_t	*c = (callstate_t *) data;
    DB		*dbp = c->i.dbp;

    debug("%s: threadid = %lx", __FUNCTION__, pthread_self());
    if ((c->status = dbp->sync(dbp, c->i.flags)) != 0) {
	debug("%s: errno = %d", __FUNCTION__, c->status);
    }
    debug("%s: threadid = %lx done", __FUNCTION__, pthread_self());
}

static void
invoke_db_remove(void *data)
{
    callstate_t	*c = (callstate_t *) data;
    DB		*dbp = c->i.dbp;

    debug("%s: threadid = %lx", __FUNCTION__, pthread_self());
    if ((c->status = dbp->remove(dbp, c->i.file, c->i.database, c->i.flags)) != 0) {
	debug("%s: errno = %d", __FUNCTION__, c->status);
    }
    debug("%s: threadid = %lx done", __FUNCTION__, pthread_self());
}

static void
invoke_db_cursor(void *data)
{
    callstate_t	*c = (callstate_t *) data;
    DB		*dbp = c->i.dbp;

    debug("%s: threadid = %lx", __FUNCTION__, pthread_self());
    if ((c->status = dbp->cursor(dbp, c->i.txnid, (DBC **) &c->o.ptr,
				 c->i.flags)) != 0) {
	debug("%s: errno = %d", __FUNCTION__, c->status);
    }
    debug("%s: threadid = %lx done", __FUNCTION__, pthread_self());
}

static void
invoke_db_c_get(void *data)
{
    callstate_t		*c = (callstate_t *) data;
    DBC			*cursorp = c->i.cursorp;
    DBT			k, d;

    debug("%s: threadid = %lx", __FUNCTION__, pthread_self());
    memset(&k, 0, sizeof(k));
    memset(&d, 0, sizeof(d));
    k.data = c->i.key;
    k.size = c->i.keylen;
    k.flags = DB_DBT_MALLOC;
    k.data = c->i.val;
    d.size = c->i.vallen;
    d.flags = DB_DBT_MALLOC;
    if ((c->status = cursorp->c_get(cursorp, &k, &d, c->i.flags)) != 0) {
	debug("%s: errno = %d", __FUNCTION__, c->status);
    } else {
	if (k.data == c->i.key) {
	    if ((c->o.key = driver_alloc_wrapper(k.size)) == NULL) {
		debug("%s: driver_alloc_wrapper failed", __FUNCTION__);
	    }
	    memcpy(c->o.key, k.data, k.size);
	} else {
	    /* XXX Is this ever the case? */
	    c->o.key = k.data;
	}
	c->o.keylen = k.size;
	c->o.val = d.data;
	c->o.vallen = d.size;
	debug("XXX %s: key = %.*s", __FUNCTION__, k.size, k.data);
	debug("XXX %s: val = %.*s", __FUNCTION__, d.size, d.data);
    }
    debug("%s: threadid = %lx done", __FUNCTION__, pthread_self());
}

static void
invoke_db_c_put(void *data)
{
    callstate_t		*c = (callstate_t *) data;
    DBC			*cursorp = c->i.cursorp;
    DBT			k, d;

    debug("%s: threadid = %lx", __FUNCTION__, pthread_self());
    memset(&k, 0, sizeof(k));
    memset(&d, 0, sizeof(d));
    k.data = c->i.key;
    k.size = c->i.keylen;
    d.data = c->i.val;
    d.size = c->i.vallen;
    if ((c->status = cursorp->c_put(cursorp, &k, &d, c->i.flags)) != 0) {
	debug("%s: errno = %d, flags = %lu", __FUNCTION__, c->status, c->i.flags);
    }
    debug("%s: threadid = %lx done", __FUNCTION__, pthread_self());
}

static void
invoke_db_c_del(void *data)
{
    callstate_t		*c = (callstate_t *) data;
    DBC			*cursorp = c->i.cursorp;

    debug("%s: threadid = %lx", __FUNCTION__, pthread_self());
    if ((c->status = cursorp->c_del(cursorp, c->i.flags)) != 0) {
	debug("%s: errno = %d", __FUNCTION__, c->status);
    }
    debug("%s: threadid = %lx done", __FUNCTION__, pthread_self());
}

static void
invoke_db_c_dup(void *data)
{
    callstate_t		*c = (callstate_t *) data;
    DBC			*cursorp = c->i.cursorp;

    debug("%s: threadid = %lx", __FUNCTION__, pthread_self());
    if ((c->status = cursorp->c_dup(cursorp, (DBC **) &c->o.ptr,
				    c->i.flags)) != 0) {
	debug("%s: errno = %d", __FUNCTION__, c->status);
    }
    debug("%s: threadid = %lx done", __FUNCTION__, pthread_self());
}

static void
invoke_db_c_count(void *data)
{
    callstate_t		*c = (callstate_t *) data;
    DBC			*cursorp = c->i.cursorp;

    debug("%s: threadid = %lx", __FUNCTION__, pthread_self());
    if ((c->status = cursorp->c_count(cursorp, &c->o.vallen, c->i.flags)) != 0) {
	debug("%s: errno = %d", __FUNCTION__, c->status);
    }
    debug("%s: threadid = %lx done", __FUNCTION__, pthread_self());
}

static void
invoke_db_c_close(void *data)
{
    callstate_t		*c = (callstate_t *) data;
    DBC			*cursorp = c->i.cursorp;

    debug("%s: threadid = %lx", __FUNCTION__, pthread_self());
    if ((c->status = cursorp->c_close(cursorp)) != 0) {
	debug("%s: errno = %d", __FUNCTION__, c->status);
    }
    debug("%s: threadid = %lx done", __FUNCTION__, pthread_self());
}

static void
invoke_db_set_pagesize(void *data)
{
    callstate_t	*c = (callstate_t *) data;
    DB		*dbp = c->i.dbp;

    debug("%s: threadid = %lx", __FUNCTION__, pthread_self());
    if ((c->status = dbp->set_pagesize(dbp, c->i.flags)) != 0) {
	debug("%s: errno = %d", __FUNCTION__, c->status);
    }
    debug("%s: threadid = %lx done", __FUNCTION__, pthread_self());
}

static void
invoke_db_set_cachesize(void *data)
{
    callstate_t	*c = (callstate_t *) data;
    DB		*dbp = c->i.dbp;

    debug("%s: threadid = %lx", __FUNCTION__, pthread_self());
    if ((c->status = dbp->set_cachesize(dbp, c->i.keylen, c->i.vallen,
					c->i.mode)) != 0) {
	debug("%s: errno = %d", __FUNCTION__, c->status);
    }
    debug("%s: threadid = %lx done", __FUNCTION__, pthread_self());
}

static void
invoke_db_set_lorder(void *data)
{
    callstate_t	*c = (callstate_t *) data;
    DB		*dbp = c->i.dbp;

    debug("%s: threadid = %lx", __FUNCTION__, pthread_self());
    if ((c->status = dbp->set_lorder(dbp, c->i.mode)) != 0) {
	debug("%s: errno = %d", __FUNCTION__, c->status);
    }
    debug("%s: threadid = %lx done", __FUNCTION__, pthread_self());
}

static void
invoke_db_set_flags(void *data)
{
    callstate_t	*c = (callstate_t *) data;
    DB		*dbp = c->i.dbp;

    debug("%s: threadid = %lx", __FUNCTION__, pthread_self());
    if ((c->status = dbp->set_flags(dbp, c->i.flags)) != 0) {
	debug("%s: errno = %d", __FUNCTION__, c->status);
    }
    debug("%s: threadid = %lx done", __FUNCTION__, pthread_self());
}

static void
invoke_db_set_bt_minkey(void *data)
{
    callstate_t	*c = (callstate_t *) data;
    DB		*dbp = c->i.dbp;

    debug("%s: threadid = %lx", __FUNCTION__, pthread_self());
    if ((c->status = dbp->set_bt_minkey(dbp, c->i.flags)) != 0) {
	debug("%s: errno = %d", __FUNCTION__, c->status);
    }
    debug("%s: threadid = %lx done", __FUNCTION__, pthread_self());
}

static void
invoke_db_set_h_ffactor(void *data)
{
    callstate_t	*c = (callstate_t *) data;
    DB		*dbp = c->i.dbp;

    debug("%s: threadid = %lx", __FUNCTION__, pthread_self());
    if ((c->status = dbp->set_h_ffactor(dbp, c->i.flags)) != 0) {
	debug("%s: errno = %d", __FUNCTION__, c->status);
    }
    debug("%s: threadid = %lx done", __FUNCTION__, pthread_self());
}

static void
invoke_db_set_h_nelem(void *data)
{
    callstate_t	*c = (callstate_t *) data;
    DB		*dbp = c->i.dbp;

    debug("%s: threadid = %lx", __FUNCTION__, pthread_self());
    if ((c->status = dbp->set_h_nelem(dbp, c->i.flags)) != 0) {
	debug("%s: errno = %d", __FUNCTION__, c->status);
    }
    debug("%s: threadid = %lx done", __FUNCTION__, pthread_self());
}

static void
invoke_db_set_re_delim(void *data)
{
    callstate_t	*c = (callstate_t *) data;
    DB		*dbp = c->i.dbp;

    debug("%s: threadid = %lx", __FUNCTION__, pthread_self());
    if ((c->status = dbp->set_re_delim(dbp, c->i.mode)) != 0) {
	debug("%s: errno = %d", __FUNCTION__, c->status);
    }
    debug("%s: threadid = %lx done", __FUNCTION__, pthread_self());
}

static void
invoke_db_set_re_len(void *data)
{
    callstate_t	*c = (callstate_t *) data;
    DB		*dbp = c->i.dbp;

    debug("%s: threadid = %lx", __FUNCTION__, pthread_self());
    if ((c->status = dbp->set_re_len(dbp, c->i.flags)) != 0) {
	debug("%s: errno = %d", __FUNCTION__, c->status);
    }
    debug("%s: threadid = %lx done", __FUNCTION__, pthread_self());
}

static void
invoke_db_set_re_pad(void *data)
{
    callstate_t	*c = (callstate_t *) data;
    DB		*dbp = c->i.dbp;

    debug("%s: threadid = %lx", __FUNCTION__, pthread_self());
    if ((c->status = dbp->set_re_pad(dbp, c->i.mode)) != 0) {
	debug("%s: errno = %d", __FUNCTION__, c->status);
    }
    debug("%s: threadid = %lx done", __FUNCTION__, pthread_self());
}

static void
invoke_db_set_re_source(void *data)
{
    callstate_t	*c = (callstate_t *) data;
    DB		*dbp = c->i.dbp;

    debug("%s: threadid = %lx", __FUNCTION__, pthread_self());
    if ((c->status = dbp->set_re_source(dbp, c->i.file)) != 0) {
	debug("%s: errno = %d", __FUNCTION__, c->status);
    }
    debug("%s: threadid = %lx done", __FUNCTION__, pthread_self());
}

