%%%----------------------------------------------------------------------
%%% File    : berkeley_db.erl
%%% Purpose : Berkeley DB driver using R8B async-style 
%%%
%%% Copyright (c) 2002, Scott Lystig Fritchie.  All rights reserved.
%%% See the file "LICENSE" at the top of the source distribution for
%%% full license terms.
%%%
%%% Important limitations of the Berkeley DB driver:
%%%
%%% 1. The driver relies on DB version 4.0.14.
%%%
%%% 2. Each open port instance can support only one DB environment.
%%%
%%% 3. Partial record storage and retrieval (via the DB_DBT_PARTIAL flag)
%%%    is not supported.  (SLF's priority: low)
%%%
%%% 4. The current implementation avoids an extra data copy by having
%%%    DB allocate memory for retrieved data via DB_DBT_MALLOC into an
%%%    ErlDrvBinary.  However, that means using DB_ENV->set_alloc()
%%%    which requires the use of an environment which requires enabling
%%%    the memory pool (DB_INIT_MPOOL).  This is extra goop for someone
%%%    who is only interested in simple DB table access.
%%%    For now, we'll live with this limitation.
%%% 5. Secondary index tables are not supported. (SLF's priority: low)
%%%
%%% 6. Returning multiple records in a single get operation (using
%%%    DB_MULTIPLE and DB_MULTIPLE_KEY) is not supported. (SLF's
%%%    priority: low)
%%%
%%% 7. Management of transactions and cursors and the limitations on 
%%%    their use by certain Pthreads is enough to give one a headache.
%%%    :-)  So, in the interest in getting a first release out the door,
%%%    the DB transaction environment (DB_INIT_TXN flag to dbenv->open())
%%%    is not supported in the first release.  (SLF's priority: high)
%%%
%%%    As a partial workaround, the Concurrent Data Store subsystem can
%%%    be used to safely perform concurrent updates on individual tables.
%%%
%%% 8. The DB_GET_BOTH option to db_get is unsupported.  (SLF's
%%%    priority: medium)
%%%
%%% 9. Quoting from the DB Reference Guide's section "Cursor Stability":
%%%
%%%      "In the absence of locking, no guarantees are made about the stability
%%%      of cursors in different threads of control. However, the Btree, Queue
%%%      and Recno access methods guarantee that cursor operations,
%%%      interspersed with any other operation in the same thread of control
%%%      will always return keys in order and will return each non-deleted
%%%      key/data pair exactly once."
%%%
%%%    The initial implementation makes no effort to coordinate which 
%%%    Pthread handles which DB operation, so cursor stability on a busy
%%%    system will be nonexistent.  The next release, with transaction
%%%    subsystem support, will fix this problem.  (SLF's priority: high)
%%%
%%%    Hrm, the DB Reference Guide states emphatically that a cursor may
%%%    not be utilized by multile threads.  {sigh}  For this release,
%%%    all cursor functions will be performed by the main Erlang VM
%%%    thread.  This means that cursor ops could block the VM in ways that
%%%    this driver is supposed to avoid, but this is a hack.  Tough.
%%%
%%% 10. The DB replication, distributed transaction management, locking,
%%%     logging, and memory pool subsystems are not supported in this release.
%%%     (SLF's priority: low to medium)
%%%
%%% 11. The "historic UNIX" dbm, ndbm, and hsearch interfaces are not
%%%     supported.  (SLF's priority: below the floor...)
%%%
%%% 12. The DB RPC client/server mechanism should work out-of-the-box,
%%%     assuming that your DB shared library was compiled with RPC support
%%%     (using "configure --enable-rpc") ... but I haven't tested it yet.
%%%
%%%----------------------------------------------------------------------

-module(berkeley_db).
-include("berkeley_db.hrl").

%% External exports
-export([startup/0, startup/4, shutdown/1, debug/2]).
-export([env_create/2, env_open/4, env_close/2, env_remove/3]).
-export([env_set_rpc_server/5]).
-export([db_create/2, db_open/7, db_put/6, db_get/5, db_del/5, db_close/3,
	 db_upgrade/4, db_truncate/4, db_verify/6, db_sync/3, db_remove/5]).
-export([db_cursor/4, c_get/5, c_put/5, c_del/3, c_dup/3, c_count/3]).
%%% TODO: db_stat
%%% TODO: db_get ... needs to pass key arg for DB_GET_BOTH??
%%% TODO: db_join ... I don't grok in order to code correctly yet.
%%% TODO: Regression test env_open/close/remove as well as db_get!
-export([db_set_pagesize/3, db_set_cachesize/5]).
-export([db_set_pagesize/3, db_set_cachesize/5, db_set_lorder/3,
	 db_set_flags/3, db_set_bt_minkey/3, db_set_h_ffactor/3,
	 db_set_h_nelem/3, db_set_re_delim/3, db_set_re_len/3,
	 db_set_re_len/3, db_set_re_pad/3, db_set_re_source/3]).
%%%-export([txn_begin/3]).
-export([regression_test/0, dets/1, dets/2]).


%%% Due to reliance on DB_ENV->set_alloc(), this startup() func is a
%%% combination of db_env_create() and DB_ENV->open().
%%%
%%% CreateFlags = flags valid for db_env_create().
%%% DB_Home = DB home valid for DB_ENV->open().
%%% EnvCreateFlags = flags valid for DB_ENV->open().
%%% EnvCreateMode = mode valid for DB_ENV->open().

startup() ->
    berkeley_db:startup(0, "", 0, 8#644).

startup(EnvCreateFlags, DB_Home, EnvOpenFlags, EnvOpenMode)
		when integer(EnvCreateFlags), integer(EnvOpenFlags),
		integer(EnvOpenMode) ->
    {ok, Path} = load_path(),
    erl_ddll:start(),
    Home = case DB_Home of
	       [] -> ".";
	       <<>> -> ".";
	       _ -> DB_Home
	   end,
    ok = erl_ddll:load_driver(Path, ?DRV_NAME),
    case open_port({spawn, ?DRV_NAME}, []) of
	P when port(P) ->
	    %% The driver will *not* work correctly without calling
	    %% env_create(), so we'll force it to happen here.
	    ok = env_create(P, EnvCreateFlags),
	    case env_open(P, Home,
			  EnvOpenFlags bor ?DB_CREATE bor ?DB_INIT_MPOOL,
			  EnvOpenMode) of
		ok -> P;
		Err -> Err
	    end;
	Err ->
	    Err
    end.

shutdown(Port) when port(Port) ->
    env_close(Port, 0),
    erlang:port_close(Port).

debug(Port, Flags) when port(Port), integer(Flags) ->
    erlang:port_command(Port, <<?BDB_DEBUG, Flags:32>>),
    get_port_reply(Port).

env_create(Port, Flags) when port(Port), integer(Flags) ->
    erlang:port_command(Port, <<?BDB_ENV_CREATE, Flags:32>>),
    get_port_reply(Port).

env_open(Port, DB_Home, Flags, Mode) when port(Port), integer(Flags),
		integer(Mode) ->
    erlang:port_command(Port, [?BDB_ENV_OPEN, DB_Home, 0, 
			       <<Flags:32, Mode:32>>]),
    get_port_reply(Port).

env_close(Port, Flags) when port(Port), integer(Flags) ->
    erlang:port_command(Port, <<?BDB_ENV_CLOSE, Flags:32>>),
    get_port_reply(Port).

env_remove(Port, DB_Home, Flags) when port(Port), integer(Flags) ->
    erlang:port_command(Port, [?BDB_ENV_REMOVE, DB_Home, 0, <<Flags:32>>]),
    get_port_reply(Port).

env_set_rpc_server(Port, Host, ClTimeout, SvTimeout, Flags) when port(Port),
	integer(ClTimeout), integer(SvTimeout), integer(Flags) ->
    erlang:port_command(Port, [?BDB_ENV_SET_RPC_SERVER, Host, 0, 
			       <<ClTimeout:32, SvTimeout:32, Flags:32>>]),
    get_port_reply(Port).

db_create(Port, Flags) when port(Port), integer(Flags) ->
    erlang:port_command(Port, <<?BDB_DB_CREATE, Flags:32>>),
    get_port_reply(Port).

db_open(Port, Table, File, Database, Type, Flags, Mode) when port(Port),
		integer(Table), Table >= 0, Table =< ?BDB_MAX_TABLES,
		integer(Type), integer(Flags), integer(Mode) ->
    erlang:port_command(Port,
			[<<?BDB_DB_OPEN, Table:8>>, File, 0, Database, 0,
			 <<Type:32, Flags:32, Mode:32>>]),
    get_port_reply(Port).

db_put(Port, Table, Txn, Key, Value, Flags) when port(Port), 
		integer(Table), Table >= 0, Table =< ?BDB_MAX_TABLES,
		integer(Txn), Txn >= ?BDB_NULL_TXN, Txn =< ?BDB_MAX_TXNS,
		integer(Flags) ->
    KLen = io_list_len(Key),
    VLen = io_list_len(Value),
    erlang:port_command(Port,
			[<<?BDB_DB_PUT, Table:8, Txn:32, KLen:32>>, Key,
			 <<VLen:32>>, Value, <<Flags:32>>]),
    get_port_reply(Port).

db_get(Port, Table, Txn, Key, Flags) when port(Port), 
		integer(Table), Table >= 0, Table =< ?BDB_MAX_TABLES,
		integer(Txn), Txn >= ?BDB_NULL_TXN, Txn =< ?BDB_MAX_TXNS,
		integer(Flags) ->
    KLen = io_list_len(Key),
    erlang:port_command(Port,
			[<<?BDB_DB_GET, Table:8, Txn:32, KLen:32>>, Key,
			 <<Flags:32>>]),
    get_port_reply(Port).

db_del(Port, Table, Txn, Key, Flags) when port(Port), 
		integer(Table), Table >= 0, Table =< ?BDB_MAX_TABLES,
		integer(Txn), Txn >= ?BDB_NULL_TXN, Txn =< ?BDB_MAX_TXNS,
		integer(Flags) ->
    KLen = io_list_len(Key),
    erlang:port_command(Port,
			[<<?BDB_DB_DEL, Table:8, Txn:32, KLen:32>>, Key,
			 <<Flags:32>>]),
    get_port_reply(Port).

db_close(Port, Table, Flags) when port(Port), 
		integer(Table), Table >= 0, Table =< ?BDB_MAX_TABLES,
		integer(Flags) ->
    erlang:port_command(Port, <<?BDB_DB_CLOSE, Table:8, Flags:32>>),
    get_port_reply(Port).

db_upgrade(Port, Table, File, Flags) when port(Port), 
		integer(Table), Table >= 0, Table =< ?BDB_MAX_TABLES,
		integer(Flags) ->
    erlang:port_command(Port, [<<?BDB_DB_UPGRADE, Table:8>>, File, 0,
			       <<Flags:32>>]),
    get_port_reply(Port).

db_truncate(Port, Table, Txn, Flags) when port(Port), 
		integer(Table), Table >= 0, Table =< ?BDB_MAX_TABLES,
		integer(Txn), Txn >= ?BDB_NULL_TXN, Txn =< ?BDB_MAX_TXNS,
		integer(Flags) ->
    erlang:port_command(Port, <<?BDB_DB_TRUNCATE, Table:8, Txn:32, Flags:32>>),
    get_port_reply(Port).

db_verify(Port, Table, File, Database, Outfile, Flags) when port(Port), 
		integer(Table), Table >= 0, Table =< ?BDB_MAX_TABLES,
		integer(Flags) ->
    erlang:port_command(Port, [<<?BDB_DB_UPGRADE, Table:8>>, File, 0,
			       Database, 0, Outfile, 0, <<Flags:32>>]),
    get_port_reply(Port).

db_remove(Port, Table, File, Database, Flags) when port(Port), 
		integer(Table), Table >= 0, Table =< ?BDB_MAX_TABLES,
		integer(Flags) ->
    erlang:port_command(Port, [<<?BDB_DB_REMOVE, Table:8>>, File, 0,
			       Database, 0, <<Flags:32>>]),
    get_port_reply(Port).

db_sync(Port, Table, Flags) when port(Port), 
		integer(Table), Table >= 0, Table =< ?BDB_MAX_TABLES,
		integer(Flags) ->
    erlang:port_command(Port, <<?BDB_DB_SYNC, Table:8, Flags:32>>),
    get_port_reply(Port).

db_cursor(Port, Table, Txn, Flags) when port(Port), 
		integer(Table), Table >= 0, Table =< ?BDB_MAX_TABLES,
		integer(Txn), Txn >= ?BDB_NULL_TXN, Txn =< ?BDB_MAX_TXNS,
		integer(Flags) ->
    erlang:port_command(Port,
			[<<?BDB_DB_CURSOR, Table:8, Txn:32, Flags:32>>]),
    get_port_reply(Port).

c_get(Port, Cursor, Key, Value, Flags) when port(Port), 
		integer(Cursor), Cursor >= 0, Cursor =< ?BDB_MAX_CURSORS,
		integer(Flags) ->
    KLen = io_list_len(Key),
    VLen = io_list_len(Value),
    erlang:port_command(Port,
			[<<?BDB_C_GET, Cursor:8, KLen:32>>, Key,
			 <<VLen:32>>, Value, <<Flags:32>>]),
    get_port_reply(Port).

c_put(Port, Cursor, Key, Value, Flags) when port(Port), 
		integer(Cursor), Cursor >= 0, Cursor =< ?BDB_MAX_CURSORS,
		integer(Flags) ->
    KLen = io_list_len(Key),
    VLen = io_list_len(Value),
    erlang:port_command(Port,
			[<<?BDB_C_PUT, Cursor:8, KLen:32>>, Key,
			 <<VLen:32>>, Value, <<Flags:32>>]),
    get_port_reply(Port).

c_del(Port, Cursor, Flags) when port(Port), 
		integer(Cursor), Cursor >= 0, Cursor =< ?BDB_MAX_CURSORS,
		integer(Flags) ->
    erlang:port_command(Port, [<<?BDB_C_DEL, Cursor:8, Flags:32>>]),
    get_port_reply(Port).

c_dup(Port, Cursor, Flags) when port(Port), 
		integer(Cursor), Cursor >= 0, Cursor =< ?BDB_MAX_CURSORS,
		integer(Flags) ->
    erlang:port_command(Port, [<<?BDB_C_DUP, Cursor:8, Flags:32>>]),
    get_port_reply(Port).

c_count(Port, Cursor, Flags) when port(Port), 
		integer(Cursor), Cursor >= 0, Cursor =< ?BDB_MAX_CURSORS,
		integer(Flags) ->
    erlang:port_command(Port, [<<?BDB_C_COUNT, Cursor:8, Flags:32>>]),
    get_port_reply(Port).

c_close(Port, Cursor) when port(Port), 
		integer(Cursor), Cursor >= 0, Cursor =< ?BDB_MAX_CURSORS ->
    erlang:port_command(Port, [<<?BDB_C_CLOSE, Cursor:8>>]),
    get_port_reply(Port).

db_set_pagesize(Port, Table, PageSize) when port(Port), 
		integer(Table), Table >= 0, Table =< ?BDB_MAX_TABLES,
		integer(PageSize) ->
    erlang:port_command(Port, <<?BDB_DB_SET_PAGESIZE, Table:8, PageSize:32>>),
    get_port_reply(Port).

db_set_cachesize(Port, Table, GBytes, Bytes, NCache) when port(Port), 
		integer(Table), Table >= 0, Table =< ?BDB_MAX_TABLES,
		integer(GBytes), integer(Bytes), integer(NCache) ->
    erlang:port_command(Port, <<?BDB_DB_SET_CACHESIZE, Table:8, GBytes:32,
			        Bytes:32, NCache:32>>),
    get_port_reply(Port).

db_set_lorder(Port, Table, LOrder) when port(Port), 
		integer(Table), Table >= 0, Table =< ?BDB_MAX_TABLES,
		integer(LOrder) ->
    erlang:port_command(Port, <<?BDB_DB_SET_LORDER, Table:8, LOrder:32>>),
    get_port_reply(Port).

db_set_flags(Port, Table, Flags) when port(Port), 
		integer(Table), Table >= 0, Table =< ?BDB_MAX_TABLES,
		integer(Flags) ->
    erlang:port_command(Port, <<?BDB_DB_SET_FLAGS, Table:8, Flags:32>>),
    get_port_reply(Port).

db_set_bt_minkey(Port, Table, BT_Minkey) when port(Port), 
		integer(Table), Table >= 0, Table =< ?BDB_MAX_TABLES,
		integer(BT_Minkey) ->
    erlang:port_command(Port, <<?BDB_DB_SET_BT_MINKEY, Table:8, BT_Minkey:32>>),
    get_port_reply(Port).

db_set_h_ffactor(Port, Table, H_FFactor) when port(Port), 
		integer(Table), Table >= 0, Table =< ?BDB_MAX_TABLES,
		integer(H_FFactor) ->
    erlang:port_command(Port, <<?BDB_DB_SET_H_FFACTOR, Table:8, H_FFactor:32>>),
    get_port_reply(Port).

db_set_h_nelem(Port, Table, H_NElem) when port(Port), 
		integer(Table), Table >= 0, Table =< ?BDB_MAX_TABLES,
		integer(H_NElem) ->
    erlang:port_command(Port, <<?BDB_DB_SET_H_NELEM, Table:8, H_NElem:32>>),
    get_port_reply(Port).

db_set_re_delim(Port, Table, RE_Delim) when port(Port), 
		integer(Table), Table >= 0, Table =< ?BDB_MAX_TABLES,
		integer(RE_Delim) ->
    erlang:port_command(Port, <<?BDB_DB_SET_RE_DELIM, Table:8, RE_Delim:32>>),
    get_port_reply(Port).

db_set_re_len(Port, Table, RE_Len) when port(Port), 
		integer(Table), Table >= 0, Table =< ?BDB_MAX_TABLES,
		integer(RE_Len) ->
    erlang:port_command(Port, <<?BDB_DB_SET_RE_LEN, Table:8, RE_Len:32>>),
    get_port_reply(Port).

db_set_re_pad(Port, Table, RE_Pad) when port(Port), 
		integer(Table), Table >= 0, Table =< ?BDB_MAX_TABLES,
		integer(RE_Pad) ->
    erlang:port_command(Port, <<?BDB_DB_SET_RE_PAD, Table:8, RE_Pad:32>>),
    get_port_reply(Port).

db_set_re_source(Port, Table, RE_Source) when port(Port), 
		integer(Table), Table >= 0, Table =< ?BDB_MAX_TABLES ->
    erlang:port_command(Port, [<<?BDB_DB_SET_RE_SOURCE, Table:8>>,
			       RE_Source, 0]),
    get_port_reply(Port).

txn_begin(Port, ParentTxn, Flags) when port(Port), integer(Flags) ->
    erlang:port_command(Port, <<?BDB_TXN_BEGIN, ParentTxn:8, Flags:32>>),
    get_port_reply(Port).

%% [ berkeley_db:regression_test() || X <- lists:duplicate(100, x) ].

regression_test() ->
    P = berkeley_db:startup(0, "", ?DB_PRIVATE, 8#644),    
%    [test_simple(P) || X <- lists:duplicate(500,x)],
    test_simple(P),
    test_simple_2(P, 50000),			% Must be even number!
    true = berkeley_db:shutdown(P),
    io:format("\nIf you made it this far, the regression test passed.\n"),
    ok.
    
%%% asleep: an example of ref-like tagging for really async behavior
%%%
%%%asleep(Port, Sec, MSec) ->
%%%    case erlang:port_control(Port, xxx, <<Sec:32, MSec:32>>) of
%%%	% <<0:8, Tag:16>> -> XXX if set_port_control_flags() worked??
%%%	[0, Tag1, Tag2] ->
%%%	    Tag = (Tag1 bsl 8) bor Tag2,
%%%	    receive
%%%		{Port, Tag, M} -> M
%%%	    end;
%%%	XXX ->
%%%	    XXX
%%%    end.

%%%
%%% Internal functions.
%%%

load_path() ->
    File = ?DRV_NAME ++ ".so",
    case lists:filter(fun(D) ->
			      case file:read_file_info(D ++ "/" ++ File) of
				  {ok, _} -> true;
				  _ -> false
			      end
		      end, code:get_path()) of
	[Dir|_] ->
	    {ok, Dir};
	[] ->
	    io:format("Error: ~s not found in code path\n", [File]),
	    {error, enoent}
    end.

get_port_reply(Port) ->
    receive
	{Port, M} -> M;
	{Port, ok, Num} when integer(Num) -> {ok, Num};
	{Port, ok, Key, Val} -> {ok, Key, Val};
	{Port, error, Reason} -> {error, Reason}
    end.

%%% TODO: io_list_len() is an extremely useful function.  BEAM has got this
%%% implemented quite efficiently in C.  It would be *fabulous* to be able
%%% to use it from Erlang via a BIF.

io_list_len(L) -> io_list_len(L, 0).
io_list_len([H|T], N) ->
    if
        H >= 0, H =< 255 -> io_list_len(T, N+1);
        list(H) -> io_list_len(T, io_list_len(H,N));
        binary(H) -> io_list_len(T, size(H) + N);
        true -> throw({error, partial_len, N})
    end;
io_list_len(H, N) when binary(H) -> 
    size(H) + N;
io_list_len([], N) -> 
    N.

%%%
%%% Regression test helpers
%%%

test_simple(P) ->
    {ok, T} = db_create(P, 0),
    DB = "regression_simple.db",
    file:delete(DB),
    ok = db_open(P, T, DB, "", ?DB_BTREE, ?DB_CREATE, 8#644),
    ok = db_put(P, T, ?BDB_NULL_TXN, "key0", "Hello, world!", 0),
    ok = db_put(P, T, ?BDB_NULL_TXN, "key0", "Hello, world!", 0),
    {error, ?DB_KEYEXIST} = db_put(P, T, ?BDB_NULL_TXN, "key0", "Hello, world!", ?DB_NOOVERWRITE),
    ok = db_put(P, T, ?BDB_NULL_TXN, "key1", "Hello, world!  This is cool!!", ?DB_NOOVERWRITE),
    ok = db_put(P, T, ?BDB_NULL_TXN, "key2", lists:duplicate(500, "Hello, great big world!  This is just too cool!!"), 0),
    ok = db_put(P, T, ?BDB_NULL_TXN, lists:duplicate(2000, "ABigKey"), "Little value", 0),
    ok = db_put(P, T, ?BDB_NULL_TXN, "deleteme", "Short-timer", 0),
    ok = db_del(P, T, ?BDB_NULL_TXN, "deleteme", 0),
    {error, ?DB_NOTFOUND} = db_del(P, T, ?BDB_NULL_TXN, "deleteme", 0),
    R = db_get(P, T, ?BDB_NULL_TXN, "key0", 0),
    ok = db_close(P, T, 0).

test_simple_2(P, Max) ->
    {ok, T} = db_create(P, 0),
    DB = "regression_simple2.db",
    file:delete(DB),
    ok = db_open(P, T, DB, "", ?DB_BTREE, ?DB_CREATE, 8#644),

    io:format("\t\tTurning off debugging to insert ~w records\n", [Max div 2]),
    {ok, OldDebug} = debug(P, 0),
    %% Eats too much memory when trying to be paranoid about driver mem leaks:
    %% lists:map(fun (X) -> K = <<X:32>>, X2 = X * 2, V = <<X2:32>>, ok = db_put(P, T, ?BDB_NULL_TXN, K, V, 0) end, lists:seq(0, (Max - 1) * 2, 2)),
    test_simple_2_insert(P, T, Max),
    io:format("\t\tReturning driver debugging to status ~w\n", [OldDebug]),
    debug(P, OldDebug),

    %% Cursor tests
    ok = test_simple_2_cursor(P, T, Max),
    ok = test_simple_2_cursordup(P, T),
    
    ok = db_close(P, T, 0).

test_simple_2_insert(P, T, X) when X =< 0 ->
    ok;
test_simple_2_insert(P, T, X) ->
    K = <<X:32>>,
    X2 = X * 2,
    V = <<X2:32>>,
    ok = db_put(P, T, ?BDB_NULL_TXN, K, V, 0),
    test_simple_2_insert(P, T, X - 2).

test_simple_2_cursor(P, T, Max) ->
    {ok, C} = db_cursor(P, T, ?BDB_NULL_TXN, 0),
    %% Position the cursor at top of table
    {ok, K, V} = c_get(P, C, <<>>, <<>>, ?DB_FIRST),
    test_simple_2_insert_cursor(P, C, K, V),
    {ok, _, _} = c_get(P, C, <<>>, <<>>, ?DB_FIRST),
    TabSize = Max,
    TabSize = test_simple_2_cursor_itemcount(P, C, 1),
    ok = c_close(P, C).

test_simple_2_insert_cursor(P, C, <<K1:32>>, <<V1:32>>) ->
    K2 = K1 + 1,
    V2 = V1 + 1,
    ok = c_put(P, C, <<K2:32>>, <<V2:32>>, ?DB_KEYFIRST),
    case c_get(P, C, <<>>, <<>>, ?DB_NEXT) of
	{ok, NK, NV} ->
	    test_simple_2_insert_cursor(P, C, NK, NV);
	{error, ?DB_NOTFOUND} ->
	    ok;
	R ->
	    io:format("ERROR: insert_cursor: c_get gave ~w\n", [R])
    end.

test_simple_2_cursor_itemcount(P, C, N) ->
    case c_get(P, C, <<>>, <<>>, ?DB_NEXT) of
	{ok, _, _} ->
	    test_simple_2_cursor_itemcount(P, C, N + 1);
	{error, ?DB_NOTFOUND} ->
	    N;
	R ->
	    io:format("ERROR: insert_cursor: c_get gave ~w\n", [R]),
	    N
    end.

test_simple_2_cursordup(P, T) ->
    {ok, C} = db_cursor(P, T, ?BDB_NULL_TXN, 0),
    %% Position the cursor at end of table
    {ok, _, _} = c_get(P, C, <<>>, <<>>, ?DB_LAST),
    {ok, _, _} = c_get(P, C, <<>>, <<>>, ?DB_PREV),
    {ok, _, _} = c_get(P, C, <<>>, <<>>, ?DB_PREV),
    {ok, _, _} = c_get(P, C, <<>>, <<>>, ?DB_PREV),
    {ok, NewC} = c_dup(P, C, ?DB_POSITION),
    {ok, K, V} = c_get(P, C, <<>>, <<>>, ?DB_PREV),
    {ok, _, _} = c_get(P, C, <<>>, <<>>, ?DB_PREV),
    {ok, _, _} = c_get(P, C, <<>>, <<>>, ?DB_PREV),
    {ok, _, _} = c_get(P, C, <<>>, <<>>, ?DB_PREV),
    {ok, _, _} = c_get(P, C, <<>>, <<>>, ?DB_PREV),
    {ok, _, _} = c_get(P, C, <<>>, <<>>, ?DB_PREV),
    {ok, _, _} = c_get(P, C, <<>>, <<>>, ?DB_PREV),
    {ok, _, _} = c_get(P, C, <<>>, <<>>, ?DB_PREV),
    {ok, _, _} = c_get(P, C, <<>>, <<>>, ?DB_PREV),
    {ok, _, _} = c_get(P, C, <<>>, <<>>, ?DB_PREV),
    {ok, K, V} = c_get(P, NewC, <<>>, <<>>, ?DB_PREV),
    c_close(P, C),
    c_close(P, NewC),
    ok.

%%% Naively simple benchmark for dets.

dets([S, RamFile]) ->		   % erl -run berkeley_db dets N true|false
    dets(list_to_integer(S), list_to_atom(RamFile)).

dets(N, RamFile) ->
    io:format("dets starting inserting ~w objects\n", [N]),
    T = dets_test,
    file:delete(atom_to_list(T)),
    {ok, T} = dets:open_file(T, [
				 {estimated_no_objects, N},
				 {ram_file, RamFile}
				]),
    dets_insert(T, N),
    ok = dets:close(T),
    io:format("dets done inserting ~w objects\n", [N]),
    ok.
    
dets_insert(T, 0) ->
    ok;
dets_insert(T, X) ->
    K = <<X:32>>,
    X2 = X * 2,
    V = <<X2:32>>,
    ok = dets:insert(T, [{K, V}]),
    dets_insert(T, X - 1).

